import asyncio
import logging
import time
from decimal import Decimal
from itertools import chain as chain_seq
from typing import Dict, List, Optional, Set, Union

from xchainpy2_mayanode import PoolsApi as PoolsApiMaya, MimirApi as MimirApiMaya, NetworkApi as NetworkApiMaya, \
    TransactionsApi as TransactionsApiMaya, LiquidityProvidersApi as LiquidityProvidersApiMaya, \
    QueueApi as QueueApiMaya, QuoteApi as QuoteApiMaya, SaversApi as SaversApiMaya
from xchainpy2_midgard import PoolDetail, THORNameDetails
from xchainpy2_midgard.api import DefaultApi as MidgardAPI
from xchainpy2_midgard.rest import ApiException
from xchainpy2_thorchain import THOR_BLOCK_TIME_SEC
from xchainpy2_thornode import PoolsApi, MimirApi, NetworkApi, InboundAddress, TransactionsApi, LiquidityProvidersApi, \
    SaversApi, QueueApi, QuoteApi, LastBlock, LiquidityProviderSummary
from xchainpy2_utils import Asset, AssetRUNE, AssetCACAO, Chain, CryptoAmount, RUNE_DECIMAL, CACAO_DECIMAL, Amount, \
    NetworkType
from .const import Mimir, TEN_MINUTES, SAME_ASSET_EXCHANGE_RATE, USD_ASSETS
from .env import URLs
from .midgard import MidgardAPIClient
from .models import PoolCache, InboundDetailCache, NetworkValuesCache, LiquidityPool, InboundDetail, SwapOutput, \
    InboundDetails, NameCache, LastBlockCache, QueryError
from .patch_clients import request_api_with_backup_hosts
from .swap import get_swap_fee, get_swap_output, get_single_swap, get_double_swap_output, \
    get_double_swap_slip
from .thornode import THORNodeAPIClient

logger = logging.getLogger('THORChainCache')

DEFAULT_MIDGARD = MidgardAPIClient()
DEFAULT_MIDGARD.configuration.host = URLs.Midgard.MAINNET

DEFAULT_THORNODE = THORNodeAPIClient()
DEFAULT_THORNODE.configuration.host = URLs.THORNode.MAINNET


class THORChainCache:
    """
    THORChainCache is a class that provides a cache for THORChain data.
    This class provides caching for data related to liquidity pools, network constants (Mimir),
    and incoming addresses and their attributes.

    """

    def __init__(self, midgard_client: MidgardAPIClient = None,
                 thornode_client: THORNodeAPIClient = None,
                 expire_pool: float = TEN_MINUTES,
                 expire_inbound: float = TEN_MINUTES,
                 expire_network: float = TEN_MINUTES,
                 native_asset: Asset = AssetRUNE,
                 network: NetworkType = NetworkType.MAINNET,
                 stable_coins: List[Asset] = None):
        """
        Constructor for THORChainCache.
        If Midgard or THORNode clients are not provided, default clients will be used.

        :param midgard_client: Тhe Midgard API client (optional)
        :param thornode_client: THORNode API client (optional)
        :param expire_pool: Expiration time for the pool cache in seconds
        :param expire_inbound: Expiration time for the inbound details cache in seconds
        :param expire_network: Expiration time for the network values cache in seconds
        :param native_asset: Native asset for the chain (RUNE or CACAO)
        :param network: Network type (MAINNET or STAGENET)
        :param stable_coins: Stable coins for the network (optional)
        """

        if not midgard_client:
            midgard_client = MidgardAPIClient()
            midgard_client.configuration.host = (
                URLs.Midgard.MAINNET if network == NetworkType.MAINNET else URLs.Midgard.STAGENET)
        else:
            assert midgard_client.configuration.host

        if not thornode_client:
            thornode_client = THORNodeAPIClient()
            thornode_client.configuration.host = (
                URLs.THORNode.MAINNET if network == NetworkType.MAINNET else URLs.THORNode.STAGENET)
        else:
            assert thornode_client.configuration.host

        self._midgard_client = midgard_client
        self._thornode_client = thornode_client
        self._pool_cache = PoolCache(0, {})
        self._inbound_cache = InboundDetailCache(0, {})
        self._network_cache = NetworkValuesCache(0, {})
        self._name_cache = NameCache({}, {}, {})
        self.expire_pool = expire_pool
        self.expire_inbound = expire_inbound
        self.expire_network = expire_network

        self.native_asset = native_asset
        self.network = network

        self.usd_stable_coins = stable_coins or USD_ASSETS[network]

        self.midgard_api = MidgardAPI(midgard_client)

        if native_asset == AssetRUNE:
            self.t_pool_api = PoolsApi(thornode_client)
            self.mimir_api = MimirApi(thornode_client)
            self.network_api = NetworkApi(thornode_client)
            self.tx_api = TransactionsApi(thornode_client)
            self.lp_api = LiquidityProvidersApi(thornode_client)
            self.queue_api = QueueApi(thornode_client)
            self.saver_api = SaversApi(thornode_client)
            self.quote_api = QuoteApi(thornode_client)
            self.chain = Chain.THORChain
            self.native_decimals = RUNE_DECIMAL
        elif native_asset == AssetCACAO:
            self.t_pool_api = PoolsApiMaya(thornode_client)
            self.mimir_api = MimirApiMaya(thornode_client)
            self.network_api = NetworkApiMaya(thornode_client)
            self.tx_api = TransactionsApiMaya(thornode_client)
            self.lp_api = LiquidityProvidersApiMaya(thornode_client)
            self.queue_api = QueueApiMaya(thornode_client)
            self.saver_api = SaversApiMaya(thornode_client)
            self.quote_api = QuoteApiMaya(thornode_client)
            self.chain = Chain.Maya
            self.native_decimals = CACAO_DECIMAL
        else:
            raise ValueError('Invalid native asset. Must be RUNE or CACAO')

        self._last_block_cache = LastBlockCache([], 0)

    @property
    def midgard_client(self) -> MidgardAPIClient:
        """
        Returns the Midgard API client.
        """
        return self._midgard_client

    @property
    def thornode_client(self) -> THORNodeAPIClient:
        """
        Returns the THORNode API client.
        """
        return self._thornode_client

    async def close(self):
        """
        Closes
        the
        Midgard and THORNode
        clients.
        It is recommended
        to
        call
        this
        method
        before
        the
        application
        exits.

        :return: None
        """
        if self._midgard_client:
            await self._midgard_client.close()
        if self._thornode_client:
            await self._thornode_client.close()

    def is_native_asset(self, a: Asset):
        """
        Checks if the
        asset is the
        native
        asset
        of
        the
        chain.

        :param
        a: Any
        Asset
        :return: True if the
        asset is the
        native
        asset, False
        otherwise
        :rtype: bool
        """
        return a == self.native_asset

    async def get_exchange_rate(self, a_from: Asset, a_to: Asset) -> Decimal:
        """
        Returns
        the
        exchange
        rate
        between
        two
        assets
        when
        using
        selected
        AMM
        protocol.
        Attention: slippage is not taken
        into
        account! Use
        THORChainQuery.quote_swap
        for more accurate simulation.

        :param
        a_from: Source
        Asset
        :param
        a_to: Destination
        Asset
        :return: Decimal
        """
        if a_from == a_to:
            return SAME_ASSET_EXCHANGE_RATE
        elif self.is_native_asset(a_from):
            lp_to = await self.get_pool_for_asset(a_to)
            return lp_to.asset_to_rune_ratio
        elif self.is_native_asset(a_to):
            lp_from = await self.get_pool_for_asset(a_from)
            return lp_from.rune_to_asset_ratio
        else:
            lp_from = await self.get_pool_for_asset(a_from)
            lp_to = await self.get_pool_for_asset(a_to)
            return lp_from.rune_to_asset_ratio * lp_to.asset_to_rune_ratio

    async def get_pool_for_asset(self, asset: Asset) -> LiquidityPool:
        """
        Returns
        the
        liquidity
        pool
        for the given asset.
        Rune
        does
        not have
        a
        pool, because
        it is the
        collateral
        for any other asset.

        :raises
        ValueError:
        if the asset is native
        :raises
        LookupError:
        if the pool is not found
        :param
        asset: Asset
        to
        find
        pool
        with
            :return: LiquidityPool
        """
        if self.is_native_asset(asset):
            raise ValueError('Native Rune does not have a pool')
        pool = self._pool_cache.pools.get(str(asset))
        if not pool:
            raise LookupError(f'Pool for {asset} not found')
        return pool

    async def get_pools(self, forced=False) -> Dict[str, LiquidityPool]:
        """
        Returns
        cached
        liquidity
        pools
        state as a
        Dict.

        :param
        forced: Force
        reload
        data
        from the API
        bypassing
        the
        cache
        :return: Dict[str, LiquidityPool]
        """

        time_elapsed = time.monotonic() - self._pool_cache.last_refreshed
        if forced or time_elapsed > self.expire_pool:
            await self.refresh_pool_cache()
        if self._pool_cache.pools:
            return self._pool_cache.pools
        raise QueryError('Could not refresh pools')

    async def refresh_pool_cache(self) -> Dict[str, LiquidityPool]:
        """
        This
        will
        reload
        the
        pool
        details
        from both THORNode and Midgard and store
        them
        into
        the
        cache.

        :return: Dict[str, LiquidityPool]
        """

        thornode_pools, midgard_pools = await asyncio.gather(
            request_api_with_backup_hosts(self.t_pool_api, self.t_pool_api.pools),
            request_api_with_backup_hosts(self._midgard_client, self.midgard_api.get_pools)
        )

        if not thornode_pools or not midgard_pools:
            raise QueryError('Could not refresh pools')

        thornode_pools_map = {str(p.asset): p for p in thornode_pools}
        midgard_pools_map = {str(p.asset): p for p in midgard_pools}
        all_assets = set(thornode_pools_map.keys()).union(set(midgard_pools_map.keys()))

        pool_map = {}
        midgard_pools: List[PoolDetail]

        for asset in all_assets:
            thornode_pool = thornode_pools_map.get(asset)
            midgard_pool = midgard_pools_map.get(asset)
            pool = LiquidityPool.from_pool_details(midgard_pool, thornode_pool)
            pool_map[str(pool.asset)] = pool

        self._pool_cache = PoolCache(time.monotonic(), pool_map)
        return self._pool_cache.pools

    async def refresh_inbound_cache(self) -> InboundDetails:
        """
        Refreshes the Inbound Details Cache which includes inbound address for each chain,
        current fee rate, halt state and more.
        It might be a better idea to call get_inbound_details() method that provides caching.

        :return InboundDetails
        """
        mimir, inbound_addresses = await asyncio.gather(
            request_api_with_backup_hosts(self.mimir_api, self.mimir_api.mimir),
            request_api_with_backup_hosts(self.network_api, self.network_api.inbound_addresses)
        )

        inbound_map = {}
        for inbound in inbound_addresses:
            inbound: InboundAddress

            if (
                    not inbound.chain or
                    not inbound.gas_rate or
                    not inbound.gas_rate_units or
                    not inbound.address or
                    not inbound.outbound_fee or
                    not inbound.outbound_tx_size
            ):
                raise QueryError('Missing required inbound info')

            halted = bool(
                inbound.halted or
                # is it necessary?
                mimir.get(Mimir.HALT_CHAIN_GLOBAL, False) or
                mimir.get(Mimir.halt_trading(inbound.chain), False)
            )

            halted_trading = bool(
                inbound.global_trading_paused or
                inbound.chain_trading_paused or
                # is it necessary?
                mimir.get(Mimir.HALT_TRADING, False) or
                mimir.get(Mimir.halt_trading(inbound.chain), False)
            )

            halted_lp = bool(
                inbound.chain_lp_actions_paused or
                # is it necessary?
                mimir.get(Mimir.PAUSE_LP, False) or
                mimir.get(Mimir.pause_lp(inbound.chain))
            )

            inbound_map[inbound.chain] = InboundDetail(
                chain=Chain(inbound.chain),
                address=inbound.address,
                gas_rate=int(inbound.gas_rate),
                gas_rate_units=inbound.gas_rate_units,
                outbound_fee=int(inbound.outbound_fee),
                outbound_tx_size=int(inbound.outbound_tx_size),
                halted_chain=halted,
                halted_trading=halted_trading,
                halted_lp=halted_lp,
                router=inbound.router,
                dust_threshold=int(inbound.dust_threshold),
            )

        inbound_map[Chain.THORChain.value] = InboundDetail(
            chain=Chain.THORChain,
            address='',
            router='',
            gas_rate=0,
            gas_rate_units='',
            outbound_fee=0,
            outbound_tx_size=0,
            halted_chain=False,
            halted_trading=not (mimir.get(Mimir.HALT_TRADING, False)),
            halted_lp=False,
            dust_threshold=0,
        )

        self._inbound_cache = InboundDetailCache(time.monotonic(), inbound_map)
        return self._inbound_cache.inbound_details

    async def refresh_network_values(self) -> Dict[str, int]:
        """
        Refreshes the NetworkValues Cache (Mimir and Constants combined).

        :return: Dict[str, int]
        """
        constants, mimir = await asyncio.gather(
            request_api_with_backup_hosts(self.network_api, self.network_api.constants),
            request_api_with_backup_hosts(self.mimir_api, self.mimir_api.mimir)
        )

        network_values = {}
        scope = chain_seq(constants.int_64_values.items(), mimir.items())
        for k, v in scope:
            network_values[k.upper()] = int(v)

        for k, v in constants.bool_values.items():
            network_values[k.upper()] = v == 'True'

        for k, v in constants.string_values.items():
            network_values[k.upper()] = v

        self._network_cache = NetworkValuesCache(time.monotonic(), network_values)
        return self._network_cache.network_values

    async def get_network_values(self, forced=False) -> Dict[str, int]:
        """
        Loads, caches and returns the Network values (Mimir and Constants combined).

        :param forced: force refresh
        :return:
        """
        sec_since_last_refresh = time.monotonic() - self._network_cache.last_refreshed
        if forced or sec_since_last_refresh > self.expire_network:
            await self.refresh_network_values()

        if not self._network_cache.network_values:
            raise QueryError('Could not refresh network values')

        return self._network_cache.network_values

    async def get_expected_swap_output(self, input_amount: CryptoAmount, dest_asset: Asset) -> SwapOutput:
        """
        This method does an approximate calculation of the swap output using known pool ratios and math.
        It takes into account only slippage but not the network fees.
        For a more accurate simulation, use THORChainQuery.quote_swap.

        :param input_amount: amount/asset to swap
        :param dest_asset: destination asset
        :return: SwapOutput structure
        """
        swap_output: SwapOutput
        if self.is_native_asset(input_amount.asset):
            # single swap from Rune -> asset
            pool = await self.get_pool_for_asset(dest_asset)
            swap_output = get_single_swap(input_amount, pool, False, self.native_asset)
        elif self.is_native_asset(dest_asset):
            # single swap from asset -> Rune
            pool = await self.get_pool_for_asset(input_amount.asset)
            swap_output = get_single_swap(input_amount, pool, True, self.native_asset)
        else:
            # double swap asset -> asset
            in_pool, out_pool = await asyncio.gather(
                self.get_pool_for_asset(input_amount.asset),
                self.get_pool_for_asset(dest_asset)
            )
            swap_output = await self._get_double_swap(input_amount, in_pool, out_pool)
        # Note this is needed to return a synth vs. a  native asset on swap out
        swap_output.output = CryptoAmount(swap_output.output.amount, dest_asset)
        return swap_output

    async def _get_double_swap(self, input_amount, in_pool, out_pool) -> SwapOutput:
        double_output = get_double_swap_output(input_amount, in_pool, out_pool)
        double_fee = await self._get_double_swap_fee(input_amount, in_pool, out_pool)
        double_slip = get_double_swap_slip(input_amount, in_pool, out_pool)
        return SwapOutput(
            output=double_output,
            swap_fee=double_fee,
            slip=double_slip,
        )

    async def _get_double_swap_fee(self, input_amount: CryptoAmount,
                                   pool1: LiquidityPool, pool2: LiquidityPool) -> CryptoAmount:
        """
        formula: getSwapFee1 + getSwapFee2
        """
        fee1_in_rune = get_swap_fee(input_amount, pool1, True)
        swap_output = get_swap_output(input_amount, pool1, True)
        fee2_in_asset = get_swap_fee(swap_output, pool2, False)
        fee2_in_rune = await self.convert(fee2_in_asset, self.native_asset)
        return fee1_in_rune + fee2_in_rune

    async def get_decimal_for_asset(self, asset: Asset) -> int:
        if asset == self.native_asset:
            return self.native_decimals
        else:
            pool = await self.get_pool_for_asset(asset)
            decimal = int(pool.thornode_details.decimals)
            return decimal if decimal > 0 else self.native_decimals

    async def convert(self, input_amount: CryptoAmount, out_asset: Asset) -> CryptoAmount:
        """
        Returns the exchange of a CryptoAmount to a different Asset
        Ex. convert(input:100 BUSD, outAsset: BTC) -> 0.0001234 BTC
        :param input_amount: amount/asset to convert to outAsset
        :param out_asset: the Asset you want to convert to
        :return: CryptoAmount of input
        """
        if input_amount.asset == out_asset:
            return input_amount

        exchange_rate = await self.get_exchange_rate(input_amount.asset, out_asset)
        out_decimals = await self.get_decimal_for_asset(out_asset)
        in_decimals = input_amount.amount.decimals
        base_amount_out = input_amount.amount * exchange_rate
        adjust_decimals = out_decimals - in_decimals
        base_amount_out *= Decimal(10 ** adjust_decimals)
        # noinspection PyTypeChecker
        amt = Amount.from_base(int(base_amount_out), out_decimals)
        return CryptoAmount(amt, out_asset)

    async def get_details_for_chain(self, chain: Union[str, Chain]) -> InboundDetail:
        """
        Returns the inbound details for a given chain. Results are cached.

        :param chain: Chain (enum or str)
        :return:
        """

        if isinstance(chain, Chain):
            chain = chain.value

        inbound = await self.get_inbound_details()
        if not inbound:
            raise QueryError('Could not get inbound details')
        if not inbound.get(chain.upper()):
            raise QueryError(f'Could not get inbound details for {chain}')
        return inbound[chain]

    async def get_inbound_details(self, forced=False) -> InboundDetails:
        """
        Returns the inbound details such as inbound vault address, fee rate and so on. Results are cached.

        :param forced: force refresh
        :return: inbound details (dict)
        """
        time_elapsed = time.monotonic() - self._inbound_cache.last_refreshed
        if forced or time_elapsed > self.expire_inbound:
            await self.refresh_inbound_cache()

        if self._inbound_cache:
            return self._inbound_cache.inbound_details
        else:
            raise QueryError('Could not refresh inbound cache')

    async def get_deepest_usd_pool(self) -> LiquidityPool:
        deepest_rune_depth = 0
        deepest_pool = None
        for usd_asset in self.usd_stable_coins:
            usd_pool = await self.get_pool_for_asset(usd_asset)
            if usd_pool.rune_balance.amount > deepest_rune_depth:
                deepest_rune_depth = usd_pool.rune_balance.amount
                deepest_pool = usd_pool
        if not deepest_pool:
            raise QueryError('no USD Pool found')
        return deepest_pool

    @property
    def is_thorchain(self):
        """
        Returns True if this class is configured for THORChain protocol.

        :return:
        """
        return self.native_asset == AssetRUNE

    @property
    def is_maya(self):
        """
        Returns True if this class is configured for Maya protocol.

        :return:
        """
        return self.native_asset == AssetCACAO

    def pluck_native_block_height(self, data: LastBlock) -> int:
        """
        Extracts the native block height from the LastBlock object.

        :param data: LastBlock object
        :return: int block height
        """
        key = 'thorchain' if self.is_thorchain else 'mayachain'
        return getattr(data, key)

    async def get_native_block_height(self) -> int:
        """
        Loads the native block height from the THORNode/MayaNode API.

        :return: int block height
        """
        data = await self.get_last_block()
        return self.pluck_native_block_height(data[0])

    async def get_last_block(self) -> List[LastBlock]:
        """
        Loads the last block information from the THORNode/MayaNode API.
        From LastBlock you can get the block height of the protocol and last observed height of the connected chains.

        :return: List[LastBlock]
        """
        t = time.monotonic()
        if t - self._last_block_cache.last_refreshed > THOR_BLOCK_TIME_SEC:
            last_block_obj = await self.network_api.lastblock()
            if not last_block_obj:
                raise QueryError("No last block")
            self._last_block_cache.last_refreshed = t
            self._last_block_cache.last_blocks = last_block_obj
            return last_block_obj
        else:
            return self._last_block_cache.last_blocks

    def get_rune_address(self, lp: LiquidityProviderSummary) -> str:
        """
        Plucks the RUNE address from the LiquidityProviderSummary object depending on the protocol.
        It is either 'rune_address' for THORChain or 'cacao_address' for MayaChain.

        :param lp: LiquidityProviderSummary
        :return: str
        """
        key = 'rune_address' if self.is_thorchain else 'cacao_address'
        return getattr(lp, key)

    async def get_liquidity_provider(self, asset: Union[str, Asset], address: str,
                                     height: int = 0) -> LiquidityProviderSummary:
        """
        Get the liquidity provider details for a given asset and address. Results are not cached.
        If height is not provided, the latest block height will be used.

        :param asset: Asset or str name of the pool
        :param address: Address of the liquidity provider
        :param height: Height of the block to query, if 0 then the latest block height will be used
        :return: LiquidityProviderSummary
        """

        if isinstance(asset, Asset):
            asset = str(asset)

        lps = await self.lp_api.liquidity_providers(asset, height=height)
        return next((lp for lp in lps
                     if lp.asset_address == address or
                     self.get_rune_address(lp) == address),
                    None)

    async def get_fee_rate(self, chain: Union[str, Chain]) -> int:
        """
        Returns the fee rate for a given chain from the inbound details. Results are cached.

        :param chain: Chain
        :return: Typical fee rate for the chain. sat/byte for Bitcoin, gas for Ethereum, etc.
        """

        inbound = await self.get_details_for_chain(chain)
        return int(inbound.gas_rate)

    async def get_names_by_address(self, address: str) -> Set[str]:
        """
        Look up THORNames that are associated with a wallet address. Names are cached.

        :raises ValueError: if address is not provided
        :param address: Address to look up
        :return: Set[str] - just the set of names without details (see: load_names_by_address)
        """

        if not address:
            raise ValueError('address is required')

        last_block = await self.get_native_block_height()
        self._name_cache.invalidate(last_block)

        if names := self._name_cache.address_to_name.get(address):
            return names

        try:
            thor_names = await self.midgard_api.get_thor_names_by_address(address)
            self._name_cache.address_to_name[address] = thor_names
            return thor_names
        except ApiException as e:
            if getattr(e, 'status', 0) == 404:
                self._name_cache.address_to_name[address] = set()
                return set()
            else:
                raise

    async def get_names_with_details(self, address: str) -> List[THORNameDetails]:
        """
        Look up THORNames with their details by a wallet address. Names are cached.
        But details are not cached yet.

        :param address: Address to look up
        :return: List[THORNameDetails]
        """

        names = await self.get_names_by_address(address)
        if not names:
            return []
        details = await asyncio.gather(*[self.get_name_details(name) for name in names])
        return list(details)

    async def get_name_details(self, name: str) -> Optional[THORNameDetails]:
        """
        Look up THORName details by a THORName. Details are cached.

        :param name: THORName to look up
        :return: Optional[THORNameDetails]
        """
        if not name:
            raise ValueError('name is required')
        name = name.lower()

        last_block = await self.get_native_block_height()
        self._name_cache.invalidate(last_block)

        if details := self._name_cache.name_details.get(name):
            return details

        try:
            thor_name = await self.midgard_api.get_thor_name_detail(name)
            self._name_cache.put(name, thor_name)
            return thor_name
        except ApiException as e:
            if e.status == 404:
                self._name_cache.put(name, None)
                return None
            else:
                raise



from decimal import Decimal

from xchainpy2_utils import CryptoAmount, Amount, AssetRUNE, Asset, DEFAULT_ASSET_DECIMAL
from .models import LiquidityPool, LPAmount, UnitData
from .swap import get_base_amount_with_diff_decimals, get_base_asset_decimals


def get_liquidity_units(liquidity: LPAmount, pool: LiquidityPool) -> int:
    """
    See: https://dev.thorchain.org/thorchain-dev/interface-guide/math#lp-units-add
    :param liquidity: asset amount added
    :param pool: pool depths
    :return: liquidity units - ownership of pool
    """
    base_amount_8_decimals = get_base_amount_with_diff_decimals(liquidity.asset, DEFAULT_ASSET_DECIMAL)
    P = Decimal(pool.pool.liquidity_units)
    r = liquidity.rune.amount.as_decimal
    a = base_amount_8_decimals
    R = pool.rune_balance.as_decimal
    A = pool.asset_balance.as_decimal
    result = P * ((R * a) + (r * A)) / (R * A * 2)
    return result


def get_pool_share(unit_data: UnitData, pool: LiquidityPool,
                   base_asset: Asset = AssetRUNE) -> LPAmount:
    """
    Get pool share of both asset and rune
    formula: (rune * part) / total; (asset * part) / total
    :param base_asset: AssetRUNE or AssetCACAO
    :param unit_data: units for both asset and rune
    :param pool: pool that the asset is bound to
    :return: pool share of both asset and rune in percentage
    """
    units = unit_data.liquidity_units
    total = unit_data.total_units
    R = pool.rune_balance.as_decimal
    T = pool.asset_balance.as_decimal
    asset = T * units / total
    rune = R * units / total
    return LPAmount(
        rune=CryptoAmount(Amount.from_base(rune, get_base_asset_decimals(base_asset)), base_asset),
        asset=CryptoAmount(Amount.from_base(asset, pool.thornode_details.decimals), pool.asset)
    )


def get_slip_on_liquidity(stake: LPAmount, pool: LiquidityPool) -> Decimal:
    """
    Get slip percentage when adding liquidity.
    Formula: (t * R - T * r)/ (T*r + R*T).

    :param stake: the share of asset and rune added to the pool
    :param pool: Pool that the asset is attached to
    :return: returns bignumber representing a slip percentage
    """
    base_amount_8_decimals = get_base_amount_with_diff_decimals(stake.asset, DEFAULT_ASSET_DECIMAL)
    r = stake.rune.amount.as_decimal
    t = base_amount_8_decimals
    R = pool.rune_balance.as_decimal
    T = pool.asset_balance.as_decimal
    numerator = (t * R) - (T * r)
    denominator = (T * r) + (R * T)
    result = numerator / denominator
    return result


def get_pool_ownership(liquidity: LPAmount, pool: LiquidityPool) -> Decimal:
    """
    Calculate liquidity units - % ownership of pool.

    See: https://docs.thorchain.org/thorchain-finance/continuous-liquidity-pools#calculating-pool-ownership
    :param liquidity: asset amount added
    :param pool: pool depths
    :return: liquidity units - % ownership of pool
    """
    P = Decimal(pool.pool.liquidity_units)
    r = liquidity.rune.amount.as_decimal
    a = liquidity.asset.amount.as_decimal
    R = pool.rune_balance.as_decimal + r  # Must add r first
    A = pool.asset_balance.as_decimal + a  # Must add t first
    part1 = R + a
    part2 = r * A
    numerator = P * (part1 + part2)
    denominator = R * A * 2
    lp_units = numerator / denominator
    percent = lp_units / P
    return percent



import asyncio
from typing import Optional

from xchainpy2_client import XChainClient
from xchainpy2_midgard.rest import ApiException
from xchainpy2_thornode import TxStatusResponse, TxSignersResponse
from xchainpy2_utils import DEFAULT_CHAIN_ATTRS, remove_0x_prefix
from .dtypes import TxDetails, TxStage, TxStatus
from ..cache import THORChainCache

DEFAULT_POLL_INTERVAL = 5  # sec
"""
Default poll interval in seconds for polling transaction status
"""


class TransactionTracker:
    def __init__(self, cache: THORChainCache, chain_attributes=None,
                 inbound_chain_client: Optional[XChainClient] = None,
                 outbound_chain_client: Optional[XChainClient] = None
                 ):
        """
        Transaction tracker.

        :param cache: THORChain cache provides access to THORChain API
        :param chain_attributes: chain attributes, don't change unless you know what you are doing
        :param inbound_chain_client: Optional, provide if you want to check tx status before it's observed by THORChain
        :param outbound_chain_client: Optional, provide if you want to check tx status after it's singed by THORChain
        """

        self.cache = cache
        self.chain_attributes = chain_attributes or DEFAULT_CHAIN_ATTRS
        self.inbound_chain_client = inbound_chain_client
        self.outbound_chain_client = outbound_chain_client

    @staticmethod
    def _validate_tx_hash(tx_hash: str):
        if not isinstance(tx_hash, str):
            raise ValueError('tx_hash should be a string')

        if len(tx_hash) <= 10:
            raise ValueError('tx_hash is too short')

    async def check_tx_outside_thorchain(self, tx_hash: str, is_inbound: bool):
        """
        Check transaction status outside THORChain, namely in the inbound or outbound chain.
        :param tx_hash: Transaction hash (tx_id)
        :param is_inbound: If True, check transaction status before it's observed by THORChain, otherwise after it's signed by THORChain
        :return:
        """
        self._validate_tx_hash(tx_hash)

        if is_inbound:
            if not self.inbound_chain_client:
                raise ValueError('inbound_chain_client is not set')
            tx_details = await self.inbound_chain_client.get_transaction_data(tx_hash)
        else:
            if not self.outbound_chain_client:
                raise ValueError('outbound_chain_client is not set')
            tx_details = await self.outbound_chain_client.get_transaction_data(tx_hash)

        return TxDetails(tx_details, )

    async def tc_get_tx_details(self, inbound_tx_hash: str, height=0) -> Optional[TxSignersResponse]:
        try:
            tx_details: TxSignersResponse = await self.cache.tx_api.tx_signers(inbound_tx_hash, height=height)
            return tx_details
        except ApiException as e:
            if e.status == 404:
                return
            raise

    async def tc_get_tx_status(self, inbound_tx_hash: str, height=0) -> Optional[TxStatusResponse]:
        """
        Get transaction status from THORChain.
        https://thornode.ninerealms.com/thorchain/tx/status/inbound_tx_hash
        According to the current implementation, if the transaction is not found, it doesn't return "None"
         but something like that

        {
          "stages": {
            "inbound_observed": {
              "started": false,
              "final_count": 0,
              "completed": false
            }
          }
        }

        :param inbound_tx_hash: Tx hash without 0x prefix
        :param height: Block height, if not provided, the latest block height is used
        :return: Optional[TxStatusResponse]
        """
        try:
            tx_status: TxStatusResponse = await self.cache.tx_api.tx_status(inbound_tx_hash, height=height)
            return tx_status
        except ApiException as e:
            if e.status == 404:
                return
            raise

    async def check_tx_progress(self, inbound_tx_hash: str) -> TxDetails:
        """
        Check transaction progress.

        :param inbound_tx_hash: Transaction hash (tx_id)
        :return: Transaction details
        """
        self._validate_tx_hash(inbound_tx_hash)

        if 'dry-run' in inbound_tx_hash.lower():
            return TxDetails.dry_run_success(inbound_tx_hash)

        tx_details = await self.tc_get_tx_details(inbound_tx_hash)
        tx_status = await self.tc_get_tx_status(inbound_tx_hash)

        if tx_details is None or tx_status is None:
            return TxDetails.create_from_thorchain(tx_details, tx_status, inbound_tx_hash)

        # todo: if there is scheduled outbound tx, check its status in its own chain

        return TxDetails.create_from_thorchain(tx_details, tx_status)

    def poll(self, txid: str, interval=DEFAULT_POLL_INTERVAL, stage=True, status=True):
        """
        Poll TX status.
        Usages:

        async for details in tracker.poll(txid):
            if not details.pending:
                print(f'Finished: {details}')
                break
            else:
                print('Still pending...')

        :param status: Flag to check status of TX
        :param stage: Flag to check stage of TX
        :param txid: inbound TX hash
        :param interval: Interval between requests in seconds
        :return: async generator
        """
        return TransactionTrackerAsyncGenerator(self, interval, txid, stage, status)


class TransactionTrackerAsyncGenerator:
    def __init__(self, tracker: TransactionTracker, interval, tx_hash, stage=True, status=True):
        self.tracker = tracker
        self.interval = interval
        self.tx_hash = remove_0x_prefix(tx_hash)
        self.stage = stage
        self.status = status
        if not self.stage and not self.status:
            raise ValueError('At least one of stage or status should be True')

        self._previous_status = TxStatus.UNKNOWN
        self._previous_stage = TxStage.Unknown
        self._finished = False

    def __aiter__(self):
        return self

    async def __anext__(self) -> TxDetails:
        if self._finished:
            raise StopAsyncIteration

        while True:
            details = await self.tracker.check_tx_progress(self.tx_hash)

            something_changed = False
            if self.stage and details.stage != self._previous_stage:
                self._previous_stage = details.stage
                something_changed = True
            if self.status and details.status != self._previous_status:
                self._previous_status = details.status
                something_changed = True

            if not details.pending:
                self._finished = True
                return details

            if not something_changed:
                await asyncio.sleep(self.interval)
            else:
                return details



import json

import pytest_asyncio

from xchainpy2_thorchain_query import ConfigurationEx, MidgardAPIClient, URLs
from xchainpy2_thorchain_query.thornode import THORNodeAPIClient


def load_json(file_name):
    if not file_name.endswith('.json'):
        file_name = f'{file_name}.json'
    with open(f'./mock_data/{file_name}') as f:
        return json.load(f)


@pytest_asyncio.fixture
async def midgard_instance():
    midgard = MidgardAPIClient(ConfigurationEx.new(host=URLs.Midgard.MAINNET))
    yield midgard
    await midgard.close()


@pytest_asyncio.fixture
async def thornode_instance():
    thornode = THORNodeAPIClient(ConfigurationEx.new(host=URLs.THORNode.MAINNET))
    yield thornode
    await thornode.close()




from xchainpy2_thornode import Pool

BUSD_POOL = Pool(
    pool_units=52543071634074,
    asset='BNB.BUSD-BD1',
    balance_asset=377399468483592,
    balance_rune=250518706651581,
    pending_inbound_asset=280314005423,
    pending_inbound_rune=533139903979,
    lp_units=56086787104869,
    savers_depth=0,
    savers_units=0,
    status='Available',
    synth_mint_paused=False,
    synth_supply=47690245926711,
    synth_supply_remaining=329709222556881,
    synth_units=3543715470795,
    derived_depth_bps='',
    loan_collateral='',
    loan_cr='',
    loan_collateral_remaining=1,
    asset_tor_price=1.0,
    savers_fill_bps=3323,
    savers_capacity_remaining=3323232,
)

def test_pool_ownership():
    ...


from multiprocessing.pool import ThreadPool

import xchainpy2_thornode as thornode
from xchainpy2_utils import NINE_REALMS_CLIENT_HEADER, XCHAINPY_IDENTIFIER, DEFAULT_USER_AGENT
from .patch_clients import ConfigurationEx, RESTClientRetry, HeadersPatch


class THORNodeAPIClient(HeadersPatch, thornode.ApiClient):
    """
    THORNode API client.

    :param configuration: ConfigurationEx object
    """
    # noinspection PyMissingConstructor
    def __init__(self, configuration: ConfigurationEx = None,
                 header_name=NINE_REALMS_CLIENT_HEADER, header_value=XCHAINPY_IDENTIFIER,
                 cookie=None):
        if configuration is None:
            configuration = ConfigurationEx()
        self.configuration = configuration

        self.pool = ThreadPool()

        # Patch REST client with additional retry logic and backup hosts
        self.rest_client = None

        self.default_headers = {}
        if header_name is not None:
            self.default_headers[header_name] = header_value
        self.cookie = cookie

        # Set default User-Agent.
        self.user_agent = DEFAULT_USER_AGENT

    async def close(self):
        """
        Close the REST client.

        :return: None
        """
        if self.rest_client:
            await self.rest_client.close()

    def request(self, method, url, query_params=None, headers=None, post_params=None, body=None, _preload_content=True,
                _request_timeout=None):
        """
        Override the request method to use the RESTClientRetry class and set the request timeout.

        :param method: Method type
        :param url: URL
        :param query_params: Parameters
        :param headers: Headers
        :param post_params: POST parameters
        :param body: The body of the request
        :param _preload_content: Flag to preload content
        :param _request_timeout: Timeout for the request
        :return: Response
        :rtype: RESTResponse
        """
        if not self.rest_client:
            # it must be initialized here, because we need to be inside asyncio loop
            self.rest_client = RESTClientRetry(self.configuration)

        if _request_timeout is None:
            _request_timeout = self.configuration.timeout

        return super().request(method, url, query_params, headers, post_params, body, _preload_content,
                               _request_timeout)



import time
from dataclasses import dataclass
from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import NamedTuple, List, Dict, Optional, Set

from xchainpy2_midgard import PoolDetail, THORNameDetails
from xchainpy2_thorchain import THOR_BLOCK_TIME_SEC
from xchainpy2_thornode import Pool, LiquidityProviderSummary, QuoteFees, LastBlock, QuoteSwapResponse, \
    QuoteSaverWithdrawResponse
from xchainpy2_utils import CryptoAmount, Amount, Asset, Chain, Address, DC


class QueryError(LookupError):
    """
    Query error is raised when the module fails to query the data from THORChain.
    """
    pass


class TotalFees(NamedTuple):
    """
    A named tuple representing the total fees for a swap transaction including outbound and affiliate fees.
    """

    asset: Asset
    """Destination asset"""

    total_bps: int
    """Total basis points for the fees"""

    total_fee: int
    """Total fee in the destination asset"""

    slippage_bps: int
    """Slippage in basis points (0-10000)"""

    affiliate_fee: int
    """Affiliate fee in the destination asset"""

    liquidity_fee: int
    """Liquidity fee in the destination asset"""

    outbound_fee: int
    """Outbound fee in the destination asset"""

    @property
    def total_fee_amount(self) -> CryptoAmount:
        return CryptoAmount(Amount.from_base(self.total_fee), self.asset)

    @property
    def affiliate_fee_amount(self) -> CryptoAmount:
        return CryptoAmount(Amount.from_base(self.affiliate_fee), self.asset)

    @property
    def liquidity_fee_amount(self) -> CryptoAmount:
        return CryptoAmount(Amount.from_base(self.liquidity_fee), self.asset)

    @property
    def outbound_fee_amount(self) -> CryptoAmount:
        return CryptoAmount(Amount.from_base(self.outbound_fee), self.asset)

    @classmethod
    def zero(cls, asset: Asset) -> 'TotalFees':
        """
        Create a zero TotalFees instance.

        :param asset: Destination asset
        :return: TotalFees instance
        """
        return cls(asset, 0, 0, 0, 0, 0, 0)


class SwapEstimate(NamedTuple):
    """
    A named tuple representing an estimate for a swap transaction.
    """

    total_fees: TotalFees
    """Total fees that will be charged for the swap"""
    slip_bps: int
    """Slippage in basis points (0-10000)"""
    net_output: CryptoAmount
    """Net output amount after fees and slippage"""
    inbound_confirmation_seconds: float
    """Estimated time in seconds for the inbound transaction to be confirmed"""
    outbound_delay_seconds: float
    """Estimated time in seconds for the outbound transaction to be delayed"""
    can_swap: bool
    """Whether the swap can be executed"""
    errors: List[str]
    """List of errors, if any (in case the swap cannot be executed)"""
    recommended_min_amount_in: int
    """Recommended minimum amount to send in the swap"""
    streaming_swap_interval: int
    """Recommended interval for streaming swaps in blocks"""
    details: QuoteSwapResponse
    """Original details of the swap estimate that was returned by THORChain API"""

    @property
    def memo(self):
        return self.details.memo

    memo.__doc__ = QuoteSwapResponse.memo.__doc__

    @property
    def notes(self):
        return self.details.notes

    notes.__doc__ = QuoteSwapResponse.notes.__doc__

    @property
    def is_less_than_price_limit(self):
        """
        Check if the swap is less than the price limit. Does search for the error message in the errors list.
        :return: True if the swap is less than the price limit, False otherwise
        :rtype: bool
        """
        if self.errors:
            return any(
                'less than price limit' in e for e in self.errors
            )
        else:
            return False

    @property
    def streaming_swap_max_quantity(self) -> int:
        """
        Get the maximum amount of trades a streaming swap can do for a trade
        """
        return self.details.max_streaming_quantity


def get_rune_balance_of_node_pool(pool: Pool) -> Amount:
    """
    Helper function to get the rune/cacao balance of a pool. Just unification of the rune/cacao balance attribute.

    :param pool: Pool object
    :return: Amount of rune/cacao in the pool
    """
    balance = getattr(pool, 'balance_rune', None)
    if balance is None:
        balance = getattr(pool, 'balance_cacao', None)
    return Amount.from_base(balance)


class LiquidityPool(NamedTuple):
    """
    Compound Names Tuple class representing a liquidity pool in THORChain.
    It includes the pool details taken from the Midgard API and THORNode API
    """

    # todo: get rid of duplicate fields, make them properties

    pool: Optional[PoolDetail]
    """Pool details from Midgard API"""
    thornode_details: Optional[Pool]
    """Pool details from THORNode API"""
    asset_balance: Amount
    """Amount of asset in the pool"""
    rune_balance: Amount
    """Amount of Rune/Cacao in the pool"""
    asset: Asset
    """Collateral asset of the pool"""
    asset_string: str
    """Collateral asset as a string"""
    rune_to_asset_ratio: Decimal
    """Rune to asset ratio; asset price in Runes"""
    asset_to_rune_ratio: Decimal
    """Asset to rune ratio; Rune price in asset units"""

    AVAILABLE = 'available'
    """Pool status: available"""

    @classmethod
    def from_pool_details(cls, pool: PoolDetail, thornode_details: Pool):
        """
        Create a LiquidityPool instance from pool details from Midgard API and THORNode API.

        :param pool: Midgard pool details
        :param thornode_details: THORNode pool details
        :return: a new LiquidityPool instance
        :rtype: LiquidityPool
        """

        ab = Amount.from_base(pool.asset_depth)
        rb = Amount.from_base(pool.rune_depth)

        ab_dec = Decimal(pool.asset_depth, DC)
        rb_dec = Decimal(pool.rune_depth, DC)

        return cls(
            pool, thornode_details,
            asset_balance=ab,
            rune_balance=rb,
            asset=Asset.from_string(pool.asset),
            asset_string=pool.asset,
            rune_to_asset_ratio=rb_dec / ab_dec,
            asset_to_rune_ratio=ab_dec / rb_dec,
        )

    @classmethod
    def from_node_pool(cls, thornode_pool: Pool):
        """
        Create a LiquidityPool instance from THORNode pool details only.

        :param thornode_pool: THORNode pool details
        :type thornode_pool: Pool
        :return: a new LiquidityPool instance
        :rtype: LiquidityPool
        """
        rune_balance = get_rune_balance_of_node_pool(thornode_pool)
        asset_balance = Amount.from_base(thornode_pool.balance_asset)

        ab_dec = asset_balance.as_decimal
        rb_dec = rune_balance.as_decimal

        return cls(
            None, thornode_pool,
            asset_balance=asset_balance,
            rune_balance=rune_balance,
            asset=Asset.from_string(thornode_pool.asset),
            asset_string=thornode_pool.asset,
            rune_to_asset_ratio=rb_dec / ab_dec,
            asset_to_rune_ratio=ab_dec / rb_dec,
        )

    def is_available(self) -> bool:
        """
        Check if the pool is in available status.

        :return: True if the pool is available, False otherwise
        :rtype: bool
        """
        return self.pool.status.lower() == self.AVAILABLE


@dataclass
class PoolCache:
    """
    A dataclass representing a cache of liquidity pools.
    """

    last_refreshed: float
    """Timestamp of the last refresh"""
    pools: Dict[str, LiquidityPool]
    """Mapping of liquidity pools by pool asset string"""


@dataclass
class NameCache:
    """
    A dataclass representing a cache of THORNames.
    """

    address_to_name: Dict[str, Set[str]]
    """Mapping of addresses to names"""
    name_details: Dict[str, THORNameDetails]
    """Mapping of names to name details"""
    name_last_refreshed: Dict[str, float]
    """Timestamp of the last refresh for each name"""

    def put(self, name: str, n: Optional[THORNameDetails]):
        """
        Put a THORNameDetails object into the cache.

        :param name: Name of the THORName
        :type name: str
        :param n: THORNameDetails object, None if the name is to be removed from the cache
        :type n: Optional[THORNameDetails]
        :return: None
        """
        if n is None and name:
            self.name_details.pop(name, None)
            self.name_last_refreshed[name] = time.monotonic()
            return

        for entry in n.entries:
            self.address_to_name.setdefault(entry.address, set()).add(name)
        if name:
            self.name_details[name] = n
            self.name_last_refreshed[name] = time.monotonic()

    @staticmethod
    def is_expired(n: THORNameDetails, last_block_height: int):
        """
        Check if a THORNameDetails object is expired.

        :param n: THORName details
        :type n: THORNameDetails
        :param last_block_height: Last block height to compare the expiry with
        :type last_block_height: int
        :return: True if the THORNameDetails object is expired, False otherwise
        :rtype: bool
        """
        return int(n.expire) < last_block_height

    def invalidate(self, block_height: int):
        """
        Invalidate expired THORNameDetails at the given block height.

        :param block_height: Block height to compare the expiry with
        :return: None
        """
        expired_names = [
            name for name, n in self.name_details.items() if self.is_expired(n, block_height)
        ]
        for name in expired_names:
            del self.name_details[name]
            del self.name_last_refreshed[name]

        for address, names in self.address_to_name.items():
            names.difference_update(expired_names)


@dataclass
class LastBlockCache:
    """
    A dataclass representing a cache of last blocks of THORChain and connected chains.
    """

    last_blocks: List[LastBlock]
    """List of last blocks"""
    last_refreshed: float
    """Timestamp of the last refresh"""


class InboundDetail(NamedTuple):
    """
    A named tuple representing the inbound details for a chain. It includes the chain, Asgard address,
    router, gas rate and other helpful information.
    See: https://thornode.ninerealms.com/thorchain/inbound_addresses
    """

    chain: Chain
    """Chain"""
    address: Address
    """Address of the vault"""
    router: Optional[Address]
    """Router address for EVM chains"""
    gas_rate: int
    """Gas rate"""
    gas_rate_units: str
    """Gas rate units: gwei, satsperbyte, etc."""
    outbound_tx_size: int
    """Outbound transaction size"""
    outbound_fee: int
    """Outbound fee"""
    halted_chain: bool
    """Whether the chain is halted"""
    halted_trading: bool
    """Whether trading is halted"""
    halted_lp: bool
    """Whether liquidity operations is halted"""
    dust_threshold: int
    """Dust threshold, do not send amounts below this value"""


InboundDetails = Dict[str, InboundDetail]
"""Mapping of chain names to inbound details"""


@dataclass
class InboundDetailCache:
    """
    A dataclass representing a cache of inbound details for chains.
    """

    last_refreshed: float
    """Timestamp of the last refresh"""
    inbound_details: InboundDetails
    """Mapping of chain names to inbound details"""


@dataclass
class NetworkValuesCache:
    """
    A dataclass representing a cache of network values (constants, Mimir) for THORChain.
    """

    last_refreshed: float
    """Timestamp of the last refresh"""
    network_values: Dict[str, int]
    """Mapping of network values"""


class UnitData(NamedTuple):
    """
    A named tuple representing the data of a unit of liquidity.
    """
    liquidity_units: int
    """Liquidity units"""
    total_units: int
    """Total units of pool"""


class LPAmount(NamedTuple):
    """
    A named tuple representing the amount of liquidity in a pool.
    """

    # todo: merge with LPAmountTotal
    rune: CryptoAmount
    """Rune amount"""
    asset: CryptoAmount
    """Asset amount"""

    @classmethod
    def zero(cls, asset: Asset = None) -> 'LPAmount':
        """
        Create a zero LPAmount instance.

        :param asset: Asset
        :return: LPAmount instance
        :rtype: LPAmount
        """
        asset = asset or Asset.from_string('')
        return cls(CryptoAmount.zero(asset), CryptoAmount.zero(asset))


class LPAmountTotal(NamedTuple):
    """
    A named tuple representing the total amount of liquidity in a pool.
    """

    rune: CryptoAmount
    """Rune amount"""
    asset: CryptoAmount
    """Asset amount"""
    total: CryptoAmount
    """Total amount in Rune"""

    @classmethod
    def zero(cls, asset: Asset = None) -> 'LPAmountTotal':
        """
        Create a zero LPAmountTotal instance.

        :param asset: Asset
        :return:
        :rtype: LPAmountTotal
        """
        asset = asset or Asset.from_string('')
        return cls(CryptoAmount.zero(asset), CryptoAmount.zero(asset), CryptoAmount.zero(asset))


class EstimateAddLP(NamedTuple):
    """
    A named tuple representing an estimate for adding liquidity to a pool.
    """

    asset_pool: str
    """Pool name"""
    slip_percent: float
    """Slippage percentage"""
    pool_share: LPAmount
    """Pool share"""
    lp_units: int
    """Liquidity units"""
    inbound_fees: LPAmountTotal
    """Inbound fees"""
    rune_to_asset_ratio: int
    """Rune to asset ratio"""
    estimated_wait_seconds: int
    """Estimated wait time in seconds"""
    errors: List[str]
    """List of errors, if any"""
    can_add: bool
    """Whether the liquidity can be added"""
    recommended_min_amount_in: int
    """Recommended minimum amount to send in the transaction"""


class WithdrawMode(Enum):
    """
    An enumeration representing the mode of withdrawal.
    """

    RuneOnly = 'RuneOnly'
    """Withdraw only Rune"""
    AssetOnly = 'AssetOnly'
    """Withdraw only Asset"""
    Symmetric = 'Symmetric'
    """Both Rune and Asset"""


class EstimateWithdrawLP(NamedTuple):
    """
    A named tuple representing an estimate for withdrawing liquidity from a pool.
    """
    can_withdraw: bool
    """Whether the liquidity can be withdrawn"""
    deposit_amount: CryptoAmount
    """Deposit amount"""
    asset_address: Optional[str]
    """Asset address"""
    rune_address: Optional[str]
    """Rune address"""
    slip_percent: float
    """Slippage percentage"""
    inbound_fee: LPAmountTotal
    """Inbound fee"""
    inbound_min_to_send: LPAmountTotal
    """Minimum amount to send"""
    outbound_fee: LPAmountTotal
    """Outbound fee"""
    asset_amount: CryptoAmount
    """Asset amount"""
    rune_amount: CryptoAmount
    """Rune amount"""
    lp_growth: str
    """LP growth"""
    estimated_wait_seconds: int
    """Estimated wait time in seconds"""
    asset_pool: str
    """Pool name"""
    errors: List[str]
    """List of errors, if any"""
    memo: str
    """Memo"""
    inbound_address: str
    """Inbound address"""
    mode: WithdrawMode

    @classmethod
    def make_error(cls, error, mode):
        zero = CryptoAmount.zero(Asset.from_string(''))
        return cls(
            False, zero, None, None, 0,
            LPAmountTotal.zero(), LPAmountTotal.zero(), LPAmountTotal.zero(),
            zero, zero,
            '', 0, '', [error],
            '', '', mode
        )


class SaverFees(NamedTuple):
    affiliate: CryptoAmount
    asset: Asset
    outbound: CryptoAmount


class EstimateWithdrawSaver(NamedTuple):
    """
    A named tuple representing an estimate for withdrawing liquidity from the savers vault.
    """

    expected_asset_amount: CryptoAmount
    """Expected asset amount that will be withdrawn to the user after applying fees and slippage"""

    fee: SaverFees
    """Fees for the withdrawal"""

    expiry: datetime
    """Expiry date of the withdrawal request"""

    to_address: Address
    """Address to send the withdraw request to"""

    memo: str
    """Memo string that is supposed to be sent along with the transaction to perform the withdrawal"""
    estimated_wait_time: float
    """Estimated wait time in seconds"""

    slip_basis_points: float
    """Slippage in basis points 0..10000 paid for internal swaps"""

    dust_amount: CryptoAmount
    """Dust amount, any amount below this value will be considered dust and ignored"""

    errors: List[str]
    """List of errors, if any"""

    details: Optional[QuoteSaverWithdrawResponse] = None
    """QuoteSaverWithdrawResponse received from THORChain API, may contain additional details"""

    @property
    def can_withdraw(self):
        """
        If there are no errors, the withdrawal can be made.
        :return: bool
        """
        return not self.errors

    @classmethod
    def make_error(cls, errors, asset: Asset, details=None):
        """
        Create an EstimateWithdrawSaver instance with errors list.

        :param errors: List of errors
        :param asset: Asset to withdraw
        :return: EstimateWithdrawSaver
        """
        return cls(
            CryptoAmount.zero(asset),
            SaverFees(CryptoAmount.zero(asset), asset, CryptoAmount.zero(asset)),
            datetime.now(), '', '', 0, 0,
            CryptoAmount.zero(asset),
            errors,
            details=details,
        )


class WithdrawLiquidityPosition(NamedTuple):
    asset: Asset
    percentage: Decimal
    asset_address: Optional[str] = None
    rune_address: Optional[str] = None


class LiquidityPosition(NamedTuple):
    pool_share: LPAmount
    position: LiquidityProviderSummary
    lp_growth: str


class PoolRatios(NamedTuple):
    asset_to_rune: Decimal
    rune_to_asset: Decimal


class EstimateAddSaver(NamedTuple):
    can_add_saver: bool
    asset_amount: CryptoAmount
    estimated_deposit_value: CryptoAmount
    slip_basis_points: int
    fee: SaverFees
    expiry: datetime
    to_address: str
    memo: str
    saver_cap_filled_percent: float
    estimated_wait_time: int
    errors: List[str]
    recommended_min_amount_in: int

    @classmethod
    def make_error(cls, errors, asset: Asset):
        return cls(
            False,
            CryptoAmount.zero(asset), CryptoAmount.zero(asset), 0,
            SaverFees(CryptoAmount.zero(asset), asset, CryptoAmount.zero(asset)),
            datetime.now(), '', '', 0, 0, errors, 0
        )


class SaversPosition(NamedTuple):
    """
    A named tuple representing the position of a saver in the savers vault.
    """

    deposit_value: CryptoAmount
    """Deposit value"""

    redeemable_value: CryptoAmount
    """Current redeemable value"""

    last_add_height: int
    """Last add height (THORChain blocks)"""

    saver_growth: float
    """Saver growth (percentage)"""

    errors: List[str]
    """List of errors, if any"""


class SwapOutput(NamedTuple):
    """
    A named tuple representing the output of a swap transaction.
    """
    output: CryptoAmount
    """Output amount and asset"""
    swap_fee: CryptoAmount
    """Swap fee amount"""
    slip: Decimal
    """Slippage 0..1"""


class BlockInformation(NamedTuple):
    """
    A named tuple representing inbound/outbound TX delay in blocks and seconds.
    """

    inbound_confirmation_blocks: int = 0
    """Inbound confirmation blocks"""
    inbound_confirmation_seconds: float = 0.0
    """Inbound confirmation seconds"""
    outbound_delay_blocks: int = 0
    """Outbound delay blocks"""
    outbound_delay_seconds: float = 0.0
    """Outbound delay seconds"""


class LoanOpenQuote(NamedTuple):
    """
    A named tuple representing a quote for opening a loan.
    """

    inbound_address: str
    """Vault's address to send your collateral to"""
    expected_wait_time: BlockInformation
    """Expected wait time for the transaction to be confirmed"""
    fees: QuoteFees
    """Fees for the transaction"""
    slippage_bps: int
    """Slippage in basis points"""
    router: str
    """Router address (for EVM chains)"""
    expiry: int
    """Expiry block number"""
    warning: str
    """Warning message"""
    notes: str
    """Notes and instructions"""
    dust_threshold: int
    """Dust threshold"""
    memo: str
    """Prepared memo for the transaction"""
    expected_amount_out: int
    """Expected amount out"""
    expected_collateralization_ratio: float
    """Expected collateralization ratio"""
    expected_collateral_up: int
    """Expected collateral up"""
    expected_debt_up: int
    """Expected debt up"""
    errors: List[str]
    """List of errors, if any"""
    recommended_min_amount_in: int
    """Recommended minimum amount to send in the transaction"""

    @classmethod
    def empty_with_errors(cls, errors):
        """
        Create an empty LoanOpenQuote instance with errors list.

        :param errors: List of errors
        :return: LoanOpenQuote
        """
        return cls(
            inbound_address='',
            expected_wait_time=BlockInformation(),
            fees=QuoteFees(),
            slippage_bps=-1,
            router='',
            expiry=-1,
            warning='',
            notes='',
            dust_threshold=0,
            memo='',
            expected_amount_out=0,
            expected_debt_up=0,
            expected_collateral_up=0,
            expected_collateralization_ratio=0,
            errors=errors,
            recommended_min_amount_in=0,
        )


class LoanCloseQuote(NamedTuple):
    """
    A named tuple representing a quote for closing a loan.
    """

    inbound_address: str
    """Inbound address to repay the loan"""
    expected_wait_time: BlockInformation
    """Expected wait time for the transaction to be confirmed"""
    fees: QuoteFees
    """Fees for the transaction"""
    slippage_bps: int
    """Slippage in basis points"""
    router: str
    """Router address (for EVM chains)"""
    expiry: int
    """Expiry block number"""
    warning: str
    """Warning message"""
    notes: str
    """Notes and instructions"""
    dust_threshold: int
    """Dust threshold"""
    memo: str
    """Prepared memo for the transaction"""
    expected_amount_out: int
    """Expected amount out"""
    expected_collateral_down: int
    """Expected collateral down"""
    expected_debt_down: int
    """Expected debt down"""
    errors: List[str]
    """List of errors, if any"""
    recommended_min_amount_in: int
    """Recommended minimum amount to send in the transaction"""

    @classmethod
    def empty_with_errors(cls, errors):
        """
        Create an empty LoanCloseQuote instance with errors list.

        :param errors: List of errors
        :return: LoanCloseQuote
        """
        return cls(
            inbound_address='',
            expected_wait_time=BlockInformation(),
            fees=QuoteFees(),
            slippage_bps=-1,
            router='',
            expiry=-1,
            warning='',
            notes='',
            dust_threshold=0,
            memo='',
            expected_amount_out=0,
            expected_collateral_down=0,
            expected_debt_down=0,
            errors=errors,
            recommended_min_amount_in=0,
        )


class THORNameEstimate(NamedTuple):
    """
    A named tuple representing a result of simulating a THORName registration. If successful, the `can_register` field
    will be `True` and the `cost` field will contain the estimated cost of the registration. If the registration is not
    possible, the `can_register` field will be `False` and the `reason` field will contain the reason why the
    registration is not possible.
    """
    can_register: bool
    """Whether the registration is possible"""
    reason: str
    """The reason why the registration is not possible"""
    cost: CryptoAmount
    """The estimated cost of the registration"""
    details: Optional[THORNameDetails] = None
    """The details of the registration, if successful"""
    last_block_number: int = 0
    """The last block number when the estimate was made"""

    def expiry_block_from_date(self, expiry: datetime) -> int:
        """
        Calculate the block number at which the registration will expire.

        :param expiry: The expiry date of the registration
        :type expiry: datetime
        :return: block number at which the registration will expire
        :rtype: int
        """
        return (expiry - datetime.now()).total_seconds() / THOR_BLOCK_TIME_SEC + self.last_block_number

    @classmethod
    def error(cls, reason: str, last_block_number: int) -> 'THORNameEstimate':
        """
        Create a THORNameEstimate instance representing an error.

        :param reason: Reason for the error
        :param last_block_number: Block number when the estimate was made
        :return: THORNameEstimate
        """
        return cls(False, reason, CryptoAmount.zero(Asset.from_string('')), None, last_block_number)



import time

import pytest
from aioresponses import aioresponses

from xchainpy2_thorchain_query import THORChainQuery
from xchainpy2_thorchain_query.tests.common import load_json


@pytest.mark.asyncio
async def test_thor_name_cached():
    with aioresponses() as m:
        m.get('https://midgard.ninerealms.com/v2/thorname/lookup/t',
              status=200,
              payload=load_json('midgard_t_name.json'))
        m.get('https://thornode.ninerealms.com/thorchain/lastblock', status=200, payload=load_json('lastblock.json'))

        query = THORChainQuery()
        name = await query.cache.get_name_details('t')

        assert name
        assert next(
            n for n in name.entries if n.chain == 'THOR').address == 'thor160yye65pf9rzwrgqmtgav69n6zlsyfpgm9a7xk'
        assert int(name.expire) == 60787996

        t0 = time.monotonic()
        await query.cache.get_name_details('t')
        t1 = time.monotonic()
        assert t1 - t0 < 0.01  # cached

        await query.close()


@pytest.mark.asyncio
async def test_thor_name_not_found():
    with aioresponses() as m:
        m.get('https://midgard.ninerealms.com/v2/thorname/lookup/tffeefffefeeefefe',
              status=404, body='not found')
        m.get('https://thornode.ninerealms.com/thorchain/lastblock', status=200, payload=load_json('lastblock.json'))

        query = THORChainQuery()
        name = await query.cache.get_name_details('tffeefffefeeefefe')
        assert not name

        await query.close()



from xchainpy2_utils import NetworkType, Asset

DEFAULT_INTERFACE_ID = 'XChainPy2'
"""
    Default Interface ID. This is used to identify the client when making requests to the public data provider.
    If you don't attach an interface ID, the server may reject your request.
"""

TC_RESERVE_ADDR = 'thor1dheycdevq39qlkxs2a6wuuzyn4aqxhve4qxtxt'
"""Reserve module's address for THORChain."""

TC_BOND_ADDR = 'thor17gw75axcnr8747pkanye45pnrwk7p9c3cqncsv'
"""Bond module's address for THORChain."""

TC_POOL_ADDR = 'thor1g98cy3n9mmjrpn0sxmn63lztelera37n8n67c0'
"""Pool module's address for THORChain."""

TC_STANDBY_RESERVE_ADDR = 'thor1lj62pg6ryxv2htekqx04nv7wd3g98qf9gfvamy'
"""Standby reserve module's address for THORChain. Burnt."""


class Mimir:
    """
    This class holds some necessary Mimir constants names for THORChain.
    """

    HALT_TRADING = 'HALTTRADING'
    HALT_CHAIN_GLOBAL = 'HALTCHAINGLOBAL'
    PAUSE_LP = 'PAUSELP'
    MIN_TX_OUT_VOLUME_THRESHOLD = 'MINTXOUTVOLUMETHRESHOLD'
    MAX_TX_OUT_OFFSET = 'MAXTXOUTOFFSET'
    TX_OUT_DELAY_RATE = 'TXOUTDELAYRATE'
    FULL_IL_PROTECTION_BLOCKS = 'FULLIMPLOSSPROTECTIONBLOCKS'

    TNS_REGISTER_FEE = 'TNSREGISTERFEE'
    TNS_FEE_PER_BLOCK = 'TNSFEEPERBLOCK'

    RUNEPoolDepositMaturityBlocks = 'RUNEPoolDepositMaturityBlocks'

    @staticmethod
    def pause_lp(chain: str):
        """
        Returns the Mimir constant name for pausing liquidity provision on a chain.

        :param chain: Name of the chain
        :type chain: str
        :return: Mimir constant name
        :rtype: str
        """
        return f'PAUSELP{chain.upper()}'

    @staticmethod
    def halt_trading(chain):
        """
        Returns the Mimir constant name for halting trading on a chain.

        :param chain: Name of the chain
        :type chain: str
        :return: Mimir constant name
        :rtype: str
        """
        return f'HALT{chain.upper()}TRADING'


DEFAULT_EXTRA_ADD_MINUTES = 15
"""Default extra minutes to add to the current time for the deadline of a transaction."""

SAME_ASSET_EXCHANGE_RATE = 1.0

TEN_MINUTES = 60 * 10
"""Ten minutes in seconds."""

USD_ASSETS = {
    NetworkType.MAINNET: [
        Asset.from_string('BNB.BUSD-BD1'),
        Asset.from_string('ETH.USDC-0XA0B86991C6218B36C1D19D4A2E9EB0CE3606EB48'),
        Asset.from_string('ETH.USDT-0XDAC17F958D2EE523A2206206994597C13D831EC7'),
        Asset.from_string('ETH.DAI-0X6B175474E89094C44DA98B954EEDEAC495271D0F'),
        Asset.from_string('ETH.GUSD-0X056FD409E1D7A124BD7017459DFEA2F387B6D5CD'),
        Asset.from_string('ETH.LUSD-0X5F98805A4E8BE255A32880FDEC7F6728C6568BA0'),
        Asset.from_string('ETH.USDP-0X8E870D67F660D95D5BE530380D0EC0BD388289E1'),
        Asset.from_string('AVAX.USDC-0XB97EF9EF8734C71904D8002F8B6BC66DD9C48A6E'),
        Asset.from_string('BSC.USDC-0X8AC76A51CC950D9822D68B83FE1AD97B32CD580D'),
    ],
    NetworkType.STAGENET: [
        Asset.from_string('ETH.USDT-0XDAC17F958D2EE523A2206206994597C13D831EC7')
    ],
    NetworkType.TESTNET: [
        Asset.from_string('BNB.BUSD-74E'),
        Asset.from_string('ETH.USDT-0XA3910454BF2CB59B8B3A401589A3BACC5CA42306')
    ]
}
"""
    Stable USD assets for different networks types. The list is based on the pool list from THORChain.
    Subject to change.
"""

THORNAME_BLOCKS_ONE_YEAR = 5259600
"""Number of blocks in one year for THORChain."""


THOR_BLOCK_TIME_SEC = 6.0
"""Typical time in seconds for a block to be produced in THORChain."""



from multiprocessing.pool import ThreadPool
from typing import Optional

import xchainpy2_midgard as mdg
from xchainpy2_utils import XCHAINPY_IDENTIFIER, NINE_REALMS_CLIENT_HEADER, DEFAULT_USER_AGENT
from .patch_clients import RESTClientRetry, ConfigurationEx, HeadersPatch


class MidgardAPIClient(HeadersPatch, mdg.ApiClient):
    """
    This class is a wrapper around the Midgard API client  :py:class:`xchainpy2_midgard.ApiClient`.
    """

    # noinspection PyMissingConstructor
    def __init__(self, configuration: Optional[ConfigurationEx] = None,
                 header_name=NINE_REALMS_CLIENT_HEADER,
                 header_value=XCHAINPY_IDENTIFIER,
                 cookie=None, pool_processes=None, user_agent=DEFAULT_USER_AGENT):
        if configuration is None:
            configuration = ConfigurationEx()
        self.configuration = configuration

        self.pool = ThreadPool(pool_processes)

        self.default_headers = {}
        if header_name is not None:
            self.default_headers[header_name] = header_value
        self.cookie = cookie

        # Set User-Agent.
        self.user_agent = user_agent

        # Patch REST client with additional retry logic and backup hosts
        self.rest_client = None

    async def close(self):
        """
        Close the REST client.

        :return: None
        """
        if self.rest_client:
            await self.rest_client.close()

    def request(self, method, url, query_params=None, headers=None, post_params=None, body=None, _preload_content=True,
                _request_timeout=None):
        """
        Override the request method to use the RESTClientRetry class and set the request timeout.

        :param method: Method type
        :param url: URL
        :param query_params: Parameters
        :param headers: Headers
        :param post_params: POST parameters
        :param body: The body of the request
        :param _preload_content: Flag to preload content
        :param _request_timeout: Timeout for the request
        :return: Response
        :rtype: RESTResponse
        """
        if not self.rest_client:
            # it must be initialized here, because we need to be inside asyncio loop
            self.rest_client = RESTClientRetry(self.configuration)

        if _request_timeout is None:
            _request_timeout = self.configuration.timeout

        return super().request(method, url, query_params, headers, post_params, body, _preload_content,
                               _request_timeout)



from decimal import Decimal

from xchainpy2_thorchain_query.models import InboundDetail
from xchainpy2_utils import CryptoAmount, Amount, CACAO_DECIMAL, AssetCACAO, Asset, AssetRUNE, RUNE_DECIMAL, Chain, \
    AssetBTC, AssetDOGE, AssetLTC, AssetBCH, AssetBNB, AssetATOM, get_chain_gas_asset, ETH_DECIMALS

try:
    from xchainpy2_mayachain import DEFAULT_CACAO_NETWORK_FEE
except ModuleNotFoundError:
    DEFAULT_CACAO_NETWORK_FEE = CryptoAmount(Amount.from_asset(Decimal("0.5"), CACAO_DECIMAL), AssetCACAO)

try:
    from xchainpy2_thorchain import DEFAULT_RUNE_NETWORK_FEE
except ModuleNotFoundError:
    DEFAULT_RUNE_NETWORK_FEE = CryptoAmount(Amount.from_asset(Decimal("0.02"), RUNE_DECIMAL), AssetRUNE)


def calc_network_fee(asset: Asset, inbound: InboundDetail,
                     base_asset: Asset = AssetRUNE) -> CryptoAmount:
    """
    Works out the required inbound fee based on the chain.
    https://dev.thorchain.org/thorchain-dev/thorchain-and-fees#fee-calcuation-by-chain

    :param base_asset:
    :param asset: source asset
    :param inbound: inbound detail to get gas rates
    :return: amount of network fee
    """
    if asset.synth:
        if base_asset == AssetRUNE:
            return DEFAULT_RUNE_NETWORK_FEE
        elif base_asset == AssetCACAO:
            return DEFAULT_CACAO_NETWORK_FEE
        else:
            raise ValueError("Invalid Base Asset, expected RUNE or CACAO")

    if asset.chain == Chain.Bitcoin.value:
        return CryptoAmount(Amount.from_base(inbound.gas_rate * inbound.outbound_tx_size), AssetBTC)
    elif asset.chain == Chain.BitcoinCash.value:
        return CryptoAmount(Amount.from_base(inbound.gas_rate * inbound.outbound_tx_size), AssetBCH)
    elif asset.chain == Chain.Litecoin.value:
        return CryptoAmount(Amount.from_base(inbound.gas_rate * inbound.outbound_tx_size), AssetLTC)
    elif asset.chain == Chain.Doge.value:
        return CryptoAmount(Amount.from_base(inbound.gas_rate * inbound.outbound_tx_size), AssetDOGE)
    elif asset.chain == Chain.Binance.value:
        return CryptoAmount(Amount.from_base(inbound.gas_rate), AssetBNB)
    elif Chain(asset.chain).is_evm:
        gas_asset = get_chain_gas_asset(Chain(asset.chain))
        decimals = ETH_DECIMALS
        gas_rate_in_gwei = Decimal(inbound.gas_rate)
        gas_rate_in_wei = Amount.from_base(gas_rate_in_gwei * Decimal(10 ** 9), decimals)
        if asset == gas_asset:
            return CryptoAmount(Amount.from_base(gas_rate_in_wei * 23000), gas_asset)
        else:
            return CryptoAmount(Amount.from_base(gas_rate_in_wei * 70000), gas_asset)
    elif asset.chain == Chain.Cosmos.value:
        return CryptoAmount(Amount.from_base(inbound.gas_rate), AssetATOM)
    elif asset.chain == Chain.THORChain.value:
        return DEFAULT_RUNE_NETWORK_FEE
    elif asset.chain == Chain.Maya.value:
        return DEFAULT_CACAO_NETWORK_FEE
    else:
        raise ValueError(f"Could not calculate inbound fee for {asset.chain} Chain")


def calc_outbound_fee(asset: Asset, inbound: InboundDetail, base_asset=AssetRUNE) -> CryptoAmount:
    """
    Works out the required outbound fee based on the chain.

    :param asset: Asset to send
    :param inbound: Inbound detail for specific chain
    :param base_asset: Rune or Cacao depending on the protocol
    :return: CryptoAmount of outbound fee
    """
    if asset.synth:
        if base_asset == AssetRUNE:
            return DEFAULT_RUNE_NETWORK_FEE
        elif base_asset == AssetCACAO:
            return DEFAULT_CACAO_NETWORK_FEE
        else:
            raise ValueError("Invalid Base Asset, expected RUNE or CACAO")

    if asset.chain == Chain.Bitcoin.value:
        return CryptoAmount(Amount.from_base(inbound.outbound_fee), AssetBTC)
    elif asset.chain == Chain.BitcoinCash.value:
        return CryptoAmount(Amount.from_base(inbound.outbound_fee), AssetBCH)
    elif asset.chain == Chain.Litecoin.value:
        return CryptoAmount(Amount.from_base(inbound.outbound_fee), AssetLTC)
    elif asset.chain == Chain.Doge.value:
        return CryptoAmount(Amount.from_base(inbound.outbound_fee), AssetDOGE)
    elif asset.chain == Chain.Binance.value:
        return CryptoAmount(Amount.from_base(inbound.outbound_fee), AssetBNB)
    elif Chain(asset.chain).is_evm:
        gas_asset = get_chain_gas_asset(Chain(asset.chain))
        decimals = ETH_DECIMALS
        wei = Decimal(inbound.outbound_fee) * Decimal(10 ** 9)
        return CryptoAmount(Amount.from_base(wei, decimals), gas_asset)
    elif asset.chain == Chain.Cosmos.value:
        return CryptoAmount(Amount.from_base(inbound.outbound_fee), AssetATOM)
    elif asset.chain == Chain.THORChain.value:
        return DEFAULT_RUNE_NETWORK_FEE
    elif asset.chain == Chain.Maya.value:
        return DEFAULT_CACAO_NETWORK_FEE
    else:
        raise ValueError(f"Could not calculate outbound fee for {asset.chain} chain")



import logging
import ssl
from typing import Optional

import aiohttp
import certifi
from aiohttp import ClientTimeout
from aiohttp.helpers import sentinel
from aiohttp_retry import ExponentialRetry, RetryClient

from xchainpy2_cosmos import DEFAULT_REST_USER_AGENT
from xchainpy2_midgard import Configuration
from xchainpy2_midgard.rest import RESTClientObject
from xchainpy2_utils import XCHAINPY_IDENTIFIER, NINE_REALMS_CLIENT_HEADER

DEFAULT_RETRY_ATTEMPTS = 3
"""
Default number of retry attempts for requests.
"""

DEFAULT_TIMEOUT = 300
"""
Default request timeout in seconds.
"""


class ConfigurationEx(Configuration):
    """
    ConfigurationEx object extends Configuration with additional parameters.
    timeout: request timeout in seconds
    retry_config: retry configuration (see: https://github.com/inyutin/aiohttp_retry)
    raise_for_status: if True, raise exceptions for HTTP errors
    backup_hosts: list of backup hosts to try if the main host fails

    """

    def __init__(self):
        super().__init__()
        self.timeout = DEFAULT_TIMEOUT
        self.retry_config = ExponentialRetry(attempts=DEFAULT_RETRY_ATTEMPTS)
        self.raise_for_status = False
        self.backup_hosts = []

    @classmethod
    def new(cls, host=None, api_key=None, api_key_prefix=None, username=None, password=None,
            ssl_ca_cert=None, cert_file=None, key_file=None, timeout=DEFAULT_TIMEOUT,
            retry_config=ExponentialRetry(attempts=DEFAULT_RETRY_ATTEMPTS), raise_for_status=False,
            backup_hosts=None, verify_ssl=True):
        """
        Create a new ConfigurationEx object.

        :param host: The main host URL
        :param api_key: Optional API key, not used by default in our case
        :param api_key_prefix: Optional API key prefix, not used by default in our case
        :param username: Optional username, not used by default in our case
        :param password: Optional password, not used by default in our case
        :param ssl_ca_cert: SSL CA certificate if needed
        :param cert_file: Certificate file if needed
        :param key_file: Key file if needed
        :param timeout: Timeout in seconds for requests
        :param retry_config: Retry configuration (see: https://github.com/inyutin/aiohttp_retry)
        :param raise_for_status: if True, raise exceptions for HTTP errors
        :param backup_hosts: An optional list of backup hosts to try if the main host fails
        :param verify_ssl: Flag to verify SSL certificates, default is True
        :return:
        """
        if backup_hosts is None:
            backup_hosts = []
        self = cls()
        self.host = host
        self.api_key = api_key
        self.api_key_prefix = api_key_prefix
        self.username = username
        self.password = password
        self.ssl_ca_cert = ssl_ca_cert
        self.cert_file = cert_file
        self.key_file = key_file
        self.timeout = timeout
        self.retry_config = retry_config
        self.raise_for_status = raise_for_status
        self.backup_hosts = backup_hosts
        self.verify_ssl = verify_ssl
        return self


class RESTClientRetry(RESTClientObject):
    # noinspection PyMissingConstructor
    def __init__(self, configuration: Optional[ConfigurationEx], pools_size=4, maxsize=4):
        # We rewrite the constructor completely to add retry logic and backup hosts
        # No need to call super().__init__ here, to avoid creating a useless session
        self.configuration = configuration

        # ca_certs
        if configuration.ssl_ca_cert:
            ca_certs = configuration.ssl_ca_cert
        else:
            # if not set certificate file, use Mozilla's root certificates.
            ca_certs = certifi.where()

        ssl_context = ssl.create_default_context(cafile=ca_certs)
        if configuration.cert_file:
            ssl_context.load_cert_chain(
                configuration.cert_file, keyfile=configuration.key_file
            )

        connector = aiohttp.TCPConnector(
            limit=maxsize,
            ssl_context=ssl_context,
            verify_ssl=configuration.verify_ssl
        )

        self.retry_config = configuration.retry_config

        # Retrying client
        timeout = sentinel
        if isinstance(configuration.timeout, (float, int)):
            timeout = ClientTimeout(total=configuration.timeout)
        elif isinstance(configuration.timeout, ClientTimeout):
            timeout = configuration.timeout

        self.pool_manager = RetryClient(
            raise_for_status=configuration.raise_for_status,
            retry_options=configuration.retry_config,
            connector=connector,
            timeout=timeout
        )
        # todo: move "proxy" parameter to request method

    async def close(self):
        await self.pool_manager.close()


logger = logging.getLogger('backup_hosts')


async def request_api_with_backup_hosts(api, method, *args, **kwargs):
    """
    Request API with backup hosts.
    It calls the specified method starting from the main host and then tries backup hosts if the main host fails.
    This is a helper function. You probably don't need to call it directly.

    :param api: API object to use
    :param method: Method to call
    :param args: Arguments
    :param kwargs: Keyword arguments
    :return:
    """
    configuration = getattr(api, 'configuration', None)
    if not configuration:
        configuration = api.api_client.configuration

    original_host = configuration.host

    hosts = []
    if original_host and original_host != '/':
        hosts.append(original_host)

    hosts.extend(configuration.backup_hosts)

    if not hosts:
        raise Exception('No hosts to try')

    last_exception = None
    for host in hosts:
        try:
            configuration.host = host
            if isinstance(method, str):
                method = getattr(api, method)
            return await method(*args, **kwargs)
        except Exception as e:
            logger.error(f'Host {host} failed for method {method}: {e}')
            last_exception = e
        finally:
            configuration.host = original_host

    if last_exception:
        raise Exception('All backup hosts failed') from last_exception

    return None


class HeadersPatch:
    """
    This is a trait class to patch REST clients with additional headers. Such as User-Agent and X-Chain-Client header.
    You probably don't need to use it directly.
    """

    def __init__(self):
        self.default_headers = None

    def patch_client(self, user_agent=DEFAULT_REST_USER_AGENT, identifier9r=XCHAINPY_IDENTIFIER):
        headers = self.default_headers
        headers['User-Agent'] = user_agent
        if identifier9r:
            headers[NINE_REALMS_CLIENT_HEADER] = identifier9r



from decimal import Decimal
from typing import Union

from xchainpy2_utils import Asset, AssetRUNE, AssetCACAO, CryptoAmount, Amount, RUNE_DECIMAL, CACAO_DECIMAL
from .models import LiquidityPool, SwapOutput


def get_base_amount_with_diff_decimals(amount: Union[CryptoAmount, Amount], out_decimals: int) -> Decimal:
    """
    Helper function to get convert an amount to a Decimal with different decimal places.

    :param amount: Input amount
    :param out_decimals: Number of decimal places to convert to
    :return: Decimal value
    """
    if isinstance(amount, CryptoAmount):
        amount = amount.amount
    new_amount = amount.changed_decimals(out_decimals)
    return Decimal(new_amount.internal_amount)


def get_base_asset_decimals(base_asset: Asset) -> int:
    """
    Helper function to get the base asset decimals.

    :param base_asset: RUNE or CACAO depending on the protocol used
    :return: int decimal places
    """
    if base_asset == AssetRUNE:
        return RUNE_DECIMAL
    elif base_asset == AssetCACAO:
        return CACAO_DECIMAL
    else:
        raise ValueError('Invalid base asset. Must be RUNE or CACAO')


def get_swap_fee(input_amount: CryptoAmount, pool: LiquidityPool, to_rune: bool,
                 base_asset: Asset = AssetRUNE) -> CryptoAmount:
    """
    Get the swap slip-based fee (slip * output) for a given swap.

    :param input_amount: amount to swap
    :param pool: Pool Data, RUNE and ASSET Depths
    :param to_rune: Direction of Swap. True if swapping to RUNE/CACAO.
    :param base_asset: AssetRUNE for TC, AssetCacao for Maya
    :return:
    """
    # formula: (x * x * Y) / (x + X) ^ 2
    base_decimals = get_base_asset_decimals(base_asset)
    decimal_out = pool.thornode_details.decimals
    if decimal_out == -1:
        decimal_out = input_amount.amount.decimals
    x = get_base_amount_with_diff_decimals(input_amount, base_decimals)
    X = pool.asset_balance.as_decimal if to_rune else pool.rune_balance.as_decimal
    Y = pool.rune_balance.as_decimal if not to_rune else pool.asset_balance.as_decimal
    units = base_asset if to_rune else pool.asset
    numerator = x * x * Y
    denominator = (x + X) ** 2
    result = numerator / denominator
    result8 = CryptoAmount(Amount.from_base(result), units)
    decimals = base_decimals if to_rune else decimal_out
    base_out = get_base_amount_with_diff_decimals(result8, decimals)
    swap_fee = CryptoAmount(Amount.automatic(base_out, decimals), units)
    return swap_fee


def get_swap_slip(input_amount: CryptoAmount, pool: LiquidityPool, to_rune: bool,
                  base_asset: Asset = AssetRUNE) -> Decimal:
    """
    Works out the swap slip for a given swap.

    :param input_amount: amount to swap
    :param pool: Pool Data, RUNE and ASSET Depths
    :param to_rune: Direction of Swap. True if swapping to RUNE.
    :param base_asset: AssetRUNE for TC, AssetCacao for Maya
    :return: The amount of slip. Needs to * 100 to get percentage.
    """
    # formula: (x) / (x + X)
    base_decimals = get_base_asset_decimals(base_asset)
    x = get_base_amount_with_diff_decimals(input_amount, base_decimals)
    X = pool.asset_balance.as_decimal if to_rune else pool.rune_balance.as_decimal
    result = x / (x + X)
    return result


def get_swap_output(input_amount: CryptoAmount, pool: LiquidityPool, to_rune: bool,
                    base_asset: Asset = AssetRUNE) -> CryptoAmount:
    """
    Works out locally the output amount for a given single swap (a single pool engaged).

    :param input_amount: amount to swap
    :param pool: Pool Data, RUNE and ASSET Depths
    :param to_rune: Direction of Swap. True if swapping to RUNE.
    :param base_asset: AssetRUNE for TC, AssetCacao for Maya
    :return: The output amount
    """
    # formula: (x * X * Y) / (x + X) ^ 2
    base_decimals = get_base_asset_decimals(base_asset)
    decimal_out = pool.thornode_details.decimals
    if decimal_out == -1:
        decimal_out = input_amount.amount.decimals
    x = get_base_amount_with_diff_decimals(input_amount, base_decimals)
    X = pool.asset_balance.as_decimal if to_rune else pool.rune_balance.as_decimal
    Y = pool.rune_balance.as_decimal if not to_rune else pool.asset_balance.as_decimal
    units = AssetRUNE if to_rune else pool.asset
    numerator = x * X * Y
    denominator = (x + X) ** 2
    result = numerator / denominator
    result8 = CryptoAmount(Amount.from_base(result), units)
    decimals = base_decimals if to_rune else decimal_out
    base_out = get_base_amount_with_diff_decimals(result8, decimals)
    output_amount = CryptoAmount(Amount.automatic(base_out, decimals), units)
    return output_amount


def get_single_swap(input_amount: CryptoAmount, pool: LiquidityPool, to_rune: bool,
                    base_asset: Asset = AssetRUNE) -> SwapOutput:
    """
    Calculates the output for a single swap (from Rune to Asset or Asset to Rune).

    :param input_amount: amount to swap
    :param pool: Pool Data, RUNE and ASSET Depths
    :param to_rune: To Rune or not to Rune
    :param base_asset: AssetRUNE for TC, AssetCacao for Maya
    :return: swap output object - output - fee - slip
    """
    output = get_swap_output(input_amount, pool, to_rune, base_asset)
    fee = get_swap_fee(input_amount, pool, to_rune, base_asset)
    slip = get_swap_slip(input_amount, pool, to_rune, base_asset)
    swap_output = SwapOutput(output, fee, slip)
    return swap_output


def get_double_swap_slip(input_amount: CryptoAmount, pool1: LiquidityPool, pool2: LiquidityPool,
                         base_asset: Asset = AssetRUNE) -> Decimal:
    """
    Works out the slippage fee for a double swap (Asset1 of pool1 -> Asset2 of pool2).
    Assets can not be RUNE. For Rune swaps please use get_swap_slip.

    :param input_amount: CryptoAmount to swap from pool1 to pool2
    :param pool1: source LiquidityPool
    :param pool2: target LiquidityPool
    :param base_asset: AssetRUNE for TC, AssetCacao for Maya
    :return:
    """
    # formula: getSwapSlip1(input1) + getSwapSlip2(getSwapOutput1 => input2)
    swap_output = get_single_swap(input_amount, pool1, True, base_asset)
    swap_output2 = get_single_swap(swap_output.output, pool2, False, base_asset)
    result = swap_output2.slip + swap_output.slip
    return result


def get_double_swap_output(input_amount: CryptoAmount, pool1: LiquidityPool, pool2: LiquidityPool) -> CryptoAmount:
    """
    This function locally estimates the output amount of a double swap.

    :param input_amount: Input amount to swap from pool1 to pool2
    :param pool1: First pool
    :type pool1: LiquidityPool
    :param pool2: Second pool
    :type pool2: LiquidityPool
    :return: CryptoAmount of output
    """

    # formula: getSwapOutput(pool1) => getSwapOutput(pool2)
    r = get_swap_output(input_amount, pool1, True)
    output = get_swap_output(r, pool2, False)
    return output



from enum import Enum
from typing import NamedTuple, Optional

from xchainpy2_thorchain import ActionType, THORMemo
from xchainpy2_thornode import TxStatusResponse, TxSignersResponse


class TxStatus(Enum):
    """
    Transaction status enum.
    If it is pending, it means that transaction is not finished yet. But you can check its stage. See TxStage enum.
    If transaction is done, it means that it is finished successfully.
    Incomplete means that you need to submit another transaction to complete this one (usually add liquidity for
    other side).
    """

    UNKNOWN = 'unknown'
    PENDING = 'pending'
    DONE = 'done'
    REFUNDED = 'refunded'
    INCOMPLETE = 'incomplete'
    BELOW_DUST = 'dust'

    Unknown = UNKNOWN
    Pending = PENDING
    Done = DONE
    Refunded = REFUNDED
    Incomplete = INCOMPLETE
    BelowDust = BELOW_DUST

    @classmethod
    def finished(cls):
        """
        Return list of statuses that are considered finished (successful or failed)
        :return:
        """
        return cls.DONE, cls.REFUNDED, cls.BELOW_DUST

    @property
    def is_successful(self):
        return self == self.DONE

    @property
    def is_pending(self):
        return self == self.PENDING

    @property
    def is_finished(self):
        return self in self.finished()


class TxStage(Enum):
    """
    Transaction stage enum.
    """

    Unknown = 'Unknown'
    """Unknown stage. Transaction is not observed yet even by the source blockchain client."""

    InboundSubmitted = 'InboundSubmitted'
    """Transaction is submitted on the inbound chain, but not observed yet by THORChain"""

    InboundObserved = 'InboundObserved'
    """Transaction is observed by THORChain"""

    InboundConfirmationCounted = 'InboundConfirmationCounted'
    """Transaction is observed and has enough confirmations"""

    InboundFinalised = 'InboundFinalised'
    """Transaction is finalised on the inbound chain"""

    SwapFinalised = 'SwapFinalised'
    """Swap transaction is finalised"""

    OutboundDelayWait = 'OutboundDelayWait'
    """Outbound transaction is waiting for delay"""

    OutboundSigned = 'OutboundSigned'
    """Outbound transaction is signed by THORChain"""

    OutboundConfirmed = 'OutboundConfirmed'
    """Outbound transaction is confirmed in the outbound chain"""


class TxDetails(NamedTuple):
    """
    Transaction details, stage and status
    """

    txid: str
    action_type: ActionType
    status: TxStatus
    stage: TxStage
    signers: TxSignersResponse
    status_details: TxStatusResponse

    @property
    def successful(self):
        return self.status == TxStatus.DONE

    @property
    def failed(self):
        return self.status in (TxStatus.REFUNDED, TxStatus.BELOW_DUST)

    @property
    def pending(self):
        return not self.failed and not self.successful

    @classmethod
    def create_from_thorchain(cls, signers: Optional[TxSignersResponse], status_details: Optional[TxStatusResponse],
                              tx_hash=''):
        if not signers or not status_details:
            return cls(
                tx_hash, ActionType.UNKNOWN, TxStatus.UNKNOWN, TxStage.Unknown, signers, status_details
            )

        memo = THORMemo.parse_memo(status_details.tx.memo)

        stage = TxStage.Unknown
        status = TxStatus.UNKNOWN

        if cls.get_stage(status_details, 'inbound_observed'):
            stage = TxStage.InboundObserved
        if cls.get_stage(status_details, 'inbound_confirmation_counted'):
            stage = TxStage.InboundConfirmationCounted
            status = TxStatus.Pending
        if cls.get_stage(status_details, 'inbound_finalised'):
            stage = TxStage.InboundFinalised
            status = TxStatus.Pending
        if cls.get_stage(status_details, 'swap_finalised'):
            stage = TxStage.SwapFinalised
        if cls.get_stage(status_details, 'outbound_delay'):
            stage = TxStage.OutboundDelayWait
        if cls.get_stage(status_details, 'outbound_signed'):
            stage = TxStage.OutboundSigned

        if stage == TxStage.OutboundSigned:
            if memo.action in (ActionType.SWAP, ActionType.WITHDRAW, ActionType.LOAN_OPEN,
                               ActionType.UNBOND):
                status = TxStatus.Done
        elif stage == TxStage.SwapFinalised and status_details.out_txs:
            # outbound inside THORChain blockchain
            status = TxStatus.DONE
        elif stage == TxStage.InboundFinalised:
            if memo.action in (ActionType.ADD_LIQUIDITY, ActionType.BOND, ActionType.DONATE):
                status = TxStatus.Done
        elif stage == TxStage.InboundObserved or stage == TxStage.InboundFinalised:
            status = TxStatus.Pending

        refunds = cls.find_refunds(status_details)
        if refunds:
            status = TxStatus.Refunded

        # todo: handle other cases and set appropriate TxStatus

        return cls(
            status_details.tx.id,
            memo.action,
            status,
            stage,
            signers, status_details
        )

    @classmethod
    def get_stage(cls, status_details, stage, key='completed', default=None):
        stages = status_details.stages.to_dict()
        stage_desc = stages.get(stage)
        if stage_desc is None:
            return

        return stage_desc.get(key, default)

    @classmethod
    def find_refunds(cls, status_details: TxStatusResponse):
        return [
            tx for tx in status_details.out_txs
            if 'REFUND' in tx.memo.upper()
        ] if status_details.out_txs else []

    @property
    def has_refunds(self):
        return bool(self.find_refunds(self.status_details))

    @property
    def has_out_txs(self):
        return bool(self.status_details.out_txs)

    @classmethod
    def dry_run_success(cls, tx_hash: str):
        """
        Create TxDetails that represents successful dry-run transaction

        :param tx_hash: TX hash string
        :return: TxDetails
        """
        # noinspection PyTypeChecker
        return cls(
            tx_hash,
            ActionType.UNKNOWN,
            TxStatus.DONE,
            TxStage.Unknown,
            None, None
        )



import asyncio
import json
import math
from datetime import datetime, timedelta
from typing import Union, List, Optional, Tuple


from xchainpy2_client import XChainClient
from xchainpy2_thorchain import THORMemo, THOR_BASIS_POINT_MAX
from xchainpy2_thornode import QuoteSwapResponse, QueueResponse, QuoteSaverDepositResponse, QuoteFees, \
    TxStatusResponse, TxSignersResponse
from xchainpy2_utils import DEFAULT_CHAIN_ATTRS, CryptoAmount, Asset, RUNE_DECIMAL, Amount, Chain, AssetRUNE, \
    get_chain_gas_asset
from .fee import calc_network_fee, calc_outbound_fee
from .midgard import MidgardAPIClient, ConfigurationEx
from .cache import THORChainCache
from .const import DEFAULT_INTERFACE_ID, Mimir, DEFAULT_EXTRA_ADD_MINUTES, THORNAME_BLOCKS_ONE_YEAR
from .liquidity import get_liquidity_units, get_pool_share, get_slip_on_liquidity
from .models import SwapEstimate, TotalFees, LPAmount, EstimateAddLP, UnitData, LPAmountTotal, \
    LiquidityPosition, PoolRatios, EstimateWithdrawLP, \
    EstimateAddSaver, SaverFees, EstimateWithdrawSaver, SaversPosition, LoanOpenQuote, \
    BlockInformation, LoanCloseQuote, THORNameEstimate, WithdrawMode, InboundDetail
from .swap import get_base_amount_with_diff_decimals
from .thornode import THORNodeAPIClient
from .track.tracker import TransactionTracker


class THORChainQuery:
    """
    THORChain Query Class is designed to query THORChain data and provide estimates for transactions.
    It also supports caching of data through THORChainCache to reduce the number of API calls.
    """

    def __init__(self,
                 cache: THORChainCache = None,
                 chain_attributes=None,
                 interface_id=DEFAULT_INTERFACE_ID,
                 native_decimal=RUNE_DECIMAL):
        """
        Initialize THORChain Query.

        :param cache: THORChainCache instance. If not provided, a new instance will be created.
        :param chain_attributes: Dictionary of chain attributes. Default is DEFAULT_CHAIN_ATTRS.
        :param interface_id: You need this field to mark your identifier for public APIs, and avoid blocking requests
        from unknown sources. Default is DEFAULT_INTERFACE_ID.
        :param native_decimal: Decimal places for the native asset, e.g. RUNE. Default is RUNE_DECIMAL.
        """

        if not cache:
            cache = THORChainCache()

        self.cache = cache
        self.chain_attributes = chain_attributes or DEFAULT_CHAIN_ATTRS
        self.interface_id = interface_id
        self.native_decimal = native_decimal

        # todo: write some tests
        self.cache.thornode_client.patch_client(self.interface_id, self.interface_id)
        self.cache.midgard_client.patch_client(self.interface_id, self.interface_id)

    @classmethod
    def from_thornode_and_midgard(cls, thornode_url='', midgard_url='',
                                  chain_attributes=None,
                                  interface_id=DEFAULT_INTERFACE_ID,
                                  native_decimal=RUNE_DECIMAL):
        """
        Create a THORChainQuery instance using custom Thornode and Midgard URLs.

        :param thornode_url: URL for Thornode API (with port and protocol)
        :param midgard_url: URL for Midgard API (with port and protocol)
        :param native_decimal: Native asset decimal places (e.g. RUNE has 8 decimals)
        :param interface_id: Interface ID for the API client. This is used to identify your requests and avoid blocking.
        :param chain_attributes: Dict of chain attributes for the query. Default is DEFAULT_CHAIN_ATTRS.

        :return: THORChainQuery instance
        """
        midgard = MidgardAPIClient(ConfigurationEx.new(host=midgard_url)) if midgard_url else MidgardAPIClient()
        thornode = THORNodeAPIClient(ConfigurationEx.new(host=thornode_url)) if thornode_url else THORNodeAPIClient()
        cache = THORChainCache(midgard, thornode)
        return cls(
            cache=cache,
            interface_id=interface_id,
            chain_attributes=chain_attributes or DEFAULT_CHAIN_ATTRS,
            native_decimal=native_decimal
        )

    @property
    def native_chain_attributes(self):
        """
        Returns the chain attributes for the native chain (e.g. THORChain)

        :return: ChainAttributes
        """
        return self.chain_attributes[self.native_asset.chain]

    @property
    def native_asset(self):
        """
        Returns the native asset (e.g. AssetRUNE or AssetCACAO).

        :return: Asset
        """
        return self.cache.native_asset

    @property
    def native_block_time(self) -> float:
        """
        Returns the average block time for the native chain.

        :return: block time in seconds
        :rtype: float
        """
        return self.chain_attributes[Chain(self.native_asset.chain)].avg_block_time

    @staticmethod
    def abbreviate_asset_string(asset: Asset, max_length=5) -> str:
        """
        Helper method to abbreviate an asset string if it is too long.

        :param asset: Asset to abbreviate
        :param max_length: Maximum length of the asset string, default is 5
        :return: Abbreviated asset string
        :rtype: str
        """

        if asset.contract and len(asset.contract) > max_length:
            abbrev = asset.contract[:max_length]
            asset = asset._replace(contract=abbrev)
        return str(asset)

    async def close(self):
        """
        Close the THORChain Query instance and the cache.
        """
        await self.cache.close()

    async def get_tx_details(self, tx_id: str) -> TxSignersResponse:
        """
        Get transaction details and status from the protocol API. See TxSignersResponse for more details.

        :param tx_id: Transaction hash (inbound)
        :return: TxSignersResponse
        """
        return await self.cache.tx_api.tx_signers(tx_id)

    async def get_tx_status(self, tx_id: str) -> TxStatusResponse:
        """
        Get transaction status and stages from the protocol API. See TxStatusResponse for more details.

        :param tx_id: Transaction hash (inbound)
        :return: TxStatusResponse
        """
        return await self.cache.tx_api.tx_status(tx_id)

    @staticmethod
    def _prepare_args_of_quote_call(**kwargs):
        # filter out empty values from kwargs
        kwargs = {key: value for key, value in kwargs.items() if value}

        affiliate_address = kwargs.get('affiliate')
        affiliate_bps = kwargs.get('affiliate_bps')

        if not affiliate_address and affiliate_bps:
            raise ValueError('Affiliate address is required if affiliate basis points are set')

        if affiliate_address and not affiliate_bps:
            kwargs['affiliate_bps'] = 0  # explicitly set to keep track of affiliate address

        return kwargs

    async def quote_swap(self,
                         input_amount: CryptoAmount,
                         destination_address: str,
                         destination_asset: Union[Asset, str],
                         tolerance_bps: int = 0,
                         affiliate_bps=0,
                         affiliate_address='',
                         streaming_interval=0,
                         streaming_quantity=0,
                         height=0,
                         ) -> SwapEstimate:
        """
        Quote a swap transaction. This is a read-only method and does not send any transactions.

        :param input_amount: CryptoAmount - amount to swap
        :param destination_address: Destination address to receive the swapped asset
        :param destination_asset: Destination asset to swap to
        :param tolerance_bps: Basis points of slippage tolerance (0..10000)
        :param affiliate_bps: Affiliate basis points (0..10000)
        :param affiliate_address: Affiliate address
        :param streaming_interval: Streaming swap interval in blocks
        :param streaming_quantity: Streaming swap quantity
        :param height: Height to query (optional), 0 for latest
        :return: Swap estimate
        :rtype: SwapEstimate
        """
        errors = []
        from_asset = str(input_amount.asset) if isinstance(input_amount.asset, Asset) else input_amount.asset
        destination_asset = str(destination_asset) if isinstance(destination_asset, Asset) else destination_asset
        input_amount_int = get_base_amount_with_diff_decimals(input_amount.amount, self.native_decimal)
        input_amount_int = int(input_amount_int)

        try:
            kwargs = self._prepare_args_of_quote_call(
                height=height,
                from_asset=from_asset,
                to_asset=destination_asset,
                amount=input_amount_int,
                destination=destination_address,
                tolerance_bps=tolerance_bps,
                affiliate_bps=affiliate_bps,
                affiliate=affiliate_address,
                streaming_interval=streaming_interval,
                streaming_quantity=streaming_quantity,
            )

            swap_quote: QuoteSwapResponse
            swap_quote = await self.cache.quote_api.quoteswap(**kwargs)

        except ValueError:
            error, response = self._get_error_and_response_from_last_thor_response()
            if error:
                errors.append(error)

            zero = CryptoAmount(Amount.zero(), AssetRUNE)
            # noinspection PyTypeChecker
            return SwapEstimate(
                TotalFees.zero(destination_asset),
                0,
                zero,
                0, 0,
                can_swap=False,
                errors=errors,
                recommended_min_amount_in=0,
                streaming_swap_interval=0,
                details=response,
            )

        if (int(swap_quote.recommended_min_amount_in) and
                int(input_amount_int) < int(swap_quote.recommended_min_amount_in)):
            recommended_in = CryptoAmount.automatic(int(swap_quote.recommended_min_amount_in), input_amount.asset,
                                                    decimals=self.native_decimal)
            errors.append(f'Input amount {input_amount} is less than recommended min amount in '
                          f'{recommended_in}')

        fee_asset = Asset.from_string_exc(swap_quote.fees.asset)

        fees: QuoteFees = swap_quote.fees
        return SwapEstimate(
            TotalFees(
                fee_asset,
                fees.total_bps, int(fees.total), fees.slippage_bps, int(fees.affiliate), int(fees.liquidity),
                int(fees.outbound),
            ),
            slip_bps=int(swap_quote.fees.slippage_bps),
            net_output=CryptoAmount(Amount(int(swap_quote.expected_amount_out)), destination_asset),
            outbound_delay_seconds=swap_quote.outbound_delay_seconds,
            inbound_confirmation_seconds=swap_quote.inbound_confirmation_seconds,
            can_swap=not bool(errors),
            errors=errors,
            recommended_min_amount_in=int(swap_quote.recommended_min_amount_in),
            streaming_swap_interval=streaming_interval,
            details=swap_quote,
        )

    # alias for quote_swap
    estimate_swap = quote_swap
    estimate_swap.__doc__ = quote_swap.__doc__

    async def outbound_delay(self, outbound_amount: CryptoAmount) -> float:
        """
        Works out how long an outbound Tx will be held by THORChain before sending.
        See https://gitlab.com/thorchain/thornode/-/blob/develop/x/thorchain/manager_txout_current.go#L548
        This function does not guarantee the exact delay.
        Please use quote_swap for a more accurate estimate.

        :param outbound_amount: CryptoAmount  being sent.
        :return: required delay in seconds
        """
        values = await self.cache.get_network_values()

        min_tx_volume_threshold = CryptoAmount(
            Amount.from_base(values.get(Mimir.MIN_TX_OUT_VOLUME_THRESHOLD, 1), self.native_decimal),
            self.cache.native_asset
        )
        max_tx_out_offset = values.get(Mimir.MAX_TX_OUT_OFFSET, 0)
        tx_out_delay_rate = float(
            Amount.from_base(values.get(Mimir.TX_OUT_DELAY_RATE, 0), self.native_decimal)
        )

        queue: QueueResponse = await self.cache.queue_api.get_queue()
        outbound_value = CryptoAmount(
            Amount.from_base(queue.scheduled_outbound_value, self.native_decimal),
            self.cache.native_asset
        )

        # blocks required to confirm tx
        avg_block_time = self.chain_attributes[self.cache.chain].avg_block_time

        # If asset is equal to Rune set runeValue as outbound amount else set it to the asset's value in rune
        rune_value = await self.cache.convert(outbound_value, self.cache.native_asset)

        # Check rune value amount
        if rune_value.amount < min_tx_volume_threshold.amount:
            return avg_block_time

        # Add OutboundAmount in rune to the outbound queue
        outbound_amount_total = outbound_amount + rune_value

        # calculate if outboundAmountTotal is over the volume threshold
        volume_threshold = outbound_amount_total / min_tx_volume_threshold

        # check delay rate
        if tx_out_delay_rate - volume_threshold.amount.amount <= 1:
            tx_out_delay_rate = 1

        # calculate the minimum number of blocks in the future the txn has to be
        min_blocks = math.ceil(rune_value.amount.amount / tx_out_delay_rate)

        min_blocks = min(max_tx_out_offset, min_blocks)
        return avg_block_time * min_blocks

    async def _get_fees_in(self, fees: TotalFees, asset: Asset) -> TotalFees:
        """
        Convenience method to convert TotalFees to a different CryptoAmount

        TotalFees are always calculated and returned in RUNE, this method can
        be used to show the equivalent fees in another Asset Type
        :param fees: TotalFees - the fees you want to convert
        :param asset: Asset - the asset you want the fees converted to
        :return: TotalFees in asset
        """
        outbound_fee, affiliate_fee, total_fee, liq_fee = await asyncio.gather(
            self.cache.convert(fees.outbound_fee_amount, asset),
            self.cache.convert(fees.affiliate_fee_amount, asset),
            self.cache.convert(fees.total_fee_amount, asset),
            self.cache.convert(fees.liquidity_fee_amount, asset),
        )
        return TotalFees(
            asset,
            total_bps=fees.total_bps,
            total_fee=int(total_fee),
            slippage_bps=fees.slippage_bps,
            affiliate_fee=int(affiliate_fee),
            liquidity_fee=int(liq_fee),
            outbound_fee=int(outbound_fee)
        )

    async def _get_confirmation_counting(self, input_coin: CryptoAmount):
        """
        Finds the required confCount required for an inbound or outbound Tx to THORChain.
        Estimate based on Midgard data only.
        Finds the gas asset of the given asset (e.g. BUSD is on BNB),
        finds the value of asset in Gas Asset then finds the required confirmation count.
        ConfCount is then times by 6 seconds.
        See https://docs.thorchain.org/chain-clients/overview
        :param input_coin: amount/asset of the outbound amount.
        :return: time in seconds before a Tx is confirmed by THORChain
        """

        # RUNE, BNB and Synths have near instant finality, so no conf counting required. - need to make a BFT only case.
        if (input_coin.asset == self.native_asset or
                input_coin.asset.chain in (Chain.Binance, Chain.Cosmos, Chain.THORChain, Chain.Maya) or
                input_coin.asset.chain):
            return self.chain_attributes[Chain.THORChain].avg_block_time
        else:
            # Get the gas asset for the inbound.asset.chain
            gas_asset = get_chain_gas_asset(input_coin.asset.chain)

            # check for chain asset, else need to convert asset value to chain asset.
            amount_in_gas_asset = await self.cache.convert(input_coin, gas_asset)

            # find the required confs
            conf_config = self.chain_attributes[input_coin.asset.chain]
            required_confs = math.ceil(amount_in_gas_asset.amount / conf_config.block_reward)
            return required_confs * conf_config.avg_block_time

    async def estimate_add_lp(self, param: LPAmount) -> EstimateAddLP:
        """
        Estimates a liquidity position for given crypto amount value, both asymmetrical and symmetrical
        :param param: LPAmount - parameters needed for a estimated liquidity position
        :return: EstimateAddLP
        """
        errors = []

        if param.asset.asset.synth or param.rune.asset.synth:
            errors.append('you cannot add liquidity with a synth')
        if param.rune.asset != self.native_asset:
            errors.append('param.rune must be THOR.RUNE/MAYA.CACAO')

        asset_pool = await self.cache.get_pool_for_asset(param.asset.asset)

        lp_units = get_liquidity_units(param, asset_pool)
        inbound_details = await self.cache.get_inbound_details()
        unit_data = UnitData(
            liquidity_units=lp_units,
            total_units=int(asset_pool.thornode_details.pool_units),
        )

        pool_share = get_pool_share(unit_data, asset_pool)

        asset_wait_time_sec = await self._get_confirmation_counting(param.asset)
        rune_wait_time_sec = await self._get_confirmation_counting(param.rune)

        wait_time_sec = max(asset_wait_time_sec, rune_wait_time_sec)

        asset_inbound_fee = CryptoAmount(Amount.zero(), param.asset.asset)
        rune_inbound_fee = CryptoAmount(Amount.zero(), self.native_asset)

        if param.asset.amount > 0:
            asset_inbound_fee = calc_network_fee(param.asset.asset, inbound_details[param.asset.asset.chain])
            if asset_inbound_fee.amount * 3 > param.asset.amount:
                errors.append('Asset amount is less than fees (3x inbound fee)')

        if param.rune.amount > 0:
            rune_inbound_fee = calc_network_fee(self.native_asset, inbound_details[self.native_asset.chain])
            if rune_inbound_fee.amount * 3 > param.rune.amount:
                errors.append('Rune amount is less than fees (3x inbound fee)')

        asset_inbound_fee_rune = await self.cache.convert(asset_inbound_fee, self.native_asset)
        total_fees_amount = asset_inbound_fee_rune + rune_inbound_fee

        slip = get_slip_on_liquidity(LPAmount(param.asset, param.rune), asset_pool)

        return EstimateAddLP(
            asset_pool=asset_pool.pool.asset,
            slip_percent=float(slip) * 100.0,
            pool_share=pool_share,
            lp_units=Amount.from_base(lp_units),
            rune_to_asset_ratio=int(asset_pool.rune_to_asset_ratio),
            inbound_fees=LPAmountTotal(
                asset=asset_inbound_fee,
                rune=rune_inbound_fee,
                total=total_fees_amount,
            ),
            estimated_wait_seconds=wait_time_sec,
            errors=errors,
            can_add=(not errors),
            recommended_min_amount_in=0,
        )

    async def check_liquidity_position(self, asset: Asset, asset_or_rune_address: str) -> LiquidityPosition:
        """
        Checks the liquidity position of a given asset and address.

        :param asset: Asset to check
        :param asset_or_rune_address: address to check
        :return: LiquidityPosition
        """
        pool_asset = await self.cache.get_pool_for_asset(asset)
        if not pool_asset:
            raise ValueError(f"Could not find pool for asset {asset}")

        liquidity_provider = await self.cache.get_liquidity_provider(
            pool_asset.asset_string,
            asset_or_rune_address,
        )

        if not liquidity_provider:
            raise ValueError(f"Could not find liquidity provider for {asset_or_rune_address}")

        # Current block number for that chain
        last_block_obj = await self.cache.get_last_block()
        block = next((b for b in last_block_obj if b.chain == asset.chain), None)
        if not block:
            raise ValueError(f"Could not find block for chain {asset.chain}")

        # Pools total units & Lp's total units
        unit_data = UnitData(
            liquidity_units=int(liquidity_provider.units),
            total_units=int(pool_asset.thornode_details.pool_units),
        )

        current_lp = LPAmount(
            asset=Amount.from_base(liquidity_provider.asset_deposit_value),
            rune=Amount.from_base(liquidity_provider.rune_deposit_value),
        )

        pool_share = get_pool_share(unit_data, pool_asset)

        # Liquidity Unit Value Index = sprt(assetDepth * runeDepth) / Pool_Units
        # Using this formula we can work out an individual position to find LUVI and then the growth rate

        deposit_luvi = math.sqrt(
            current_lp.asset.amount * current_lp.rune.amount / unit_data.liquidity_units
        )

        redeem_luvi = math.sqrt(
            pool_share.asset.amount.amount * pool_share.rune.amount.amount / unit_data.liquidity_units
        )

        lp_growth = redeem_luvi - deposit_luvi
        current_lp_growth = (lp_growth / deposit_luvi if lp_growth > 0 else 0.0) * 100.0

        return LiquidityPosition(
            pool_share,
            liquidity_provider,
            f'{current_lp_growth:.2f} %',
        )

    async def get_pool_ratio(self, asset: Asset) -> PoolRatios:
        """
        Gets the pool ratio for a given asset
        Do not send assetNativeRune, There is no pool for it.
        :param asset: asset required to find the pool
        :return: object type ratios
        """
        asset_pool = await self.cache.get_pool_for_asset(asset)
        return PoolRatios(
            asset_to_rune=asset_pool.asset_to_rune_ratio,
            rune_to_asset=asset_pool.rune_to_asset_ratio,
        )

    async def estimate_withdraw_lp(self, asset: Asset,
                                   mode: WithdrawMode,
                                   withdraw_bps: int = THOR_BASIS_POINT_MAX,
                                   asset_address: Optional[str] = None,
                                   rune_address: Optional[str] = None,
                                   ) -> EstimateWithdrawLP:
        """
        Estimate the withdrawal of liquidity
        :param asset: Pool to withdraw from
        :param mode: withdrawal mode (asset, rune, symmetric)
        :param withdraw_bps: basis points to withdraw 0..10k (0..100%)
        :param asset_address: asset address (optional)
        :param rune_address: rune address (optional)
        :return:
        """
        asset = Asset.automatic(asset)
        if not asset.chain or not asset.symbol:
            return EstimateWithdrawLP.make_error(f"Invalid asset {asset}", mode)

        if withdraw_bps <= 0 or withdraw_bps > THOR_BASIS_POINT_MAX:
            return EstimateWithdrawLP.make_error(f"Invalid withdraw basis points {withdraw_bps}; "
                                                 f"must be 0..{THOR_BASIS_POINT_MAX}", mode)

        # Caution Dust Limits: BTC, BCH, LTC chains 10k sats; DOGE 1m Sats; ETH 0 wei; THOR 0 RUNE.
        member_address = rune_address if rune_address else asset_address

        member_detail = await self.check_liquidity_position(asset, member_address)
        if not member_detail or not member_detail.position.units:
            return EstimateWithdrawLP.make_error(f"Address {member_address} has no liquidity position for {asset}",
                                                 mode)

        dust_values = DEFAULT_CHAIN_ATTRS[asset.chain].dust
        asset_pool = await self.cache.get_pool_for_asset(asset)
        if not asset_pool:
            return EstimateWithdrawLP.make_error(f"Could not find pool for asset {asset}", mode)

        inbound_address = ''
        if not rune_address and asset_address:
            inbound_details = await self.cache.get_inbound_details()
            inbound_address = inbound_details.get(asset.chain, {}).get('address', '')
            if not inbound_address:
                return EstimateWithdrawLP.make_error(f"Could not find inbound address for {asset}", mode)

        # get pool share from unit data
        pool_share = get_pool_share(
            UnitData(
                liquidity_units=int(member_detail.position.units),
                total_units=int(asset_pool.pool.liquidity_units),
            ),
            asset_pool,
        )

        # get slip on liquidity removal
        slip = get_slip_on_liquidity(pool_share, asset_pool)

        # TODO make sure we compare wait times for withdrawing both rune and asset OR just rune OR just asset
        withdraw_part = withdraw_bps / THOR_BASIS_POINT_MAX
        wait_time_sec_for_asset, wait_time_sec_for_rune = await asyncio.gather(
            self._get_confirmation_counting(pool_share.asset / withdraw_part),
            self._get_confirmation_counting(pool_share.rune / withdraw_part)
        )

        # todo: is it cacao for maya?
        if member_detail.position.rune_address and member_detail.position.asset_address:
            wait_time_in_sec = max(wait_time_sec_for_asset, wait_time_sec_for_rune)
        elif member_detail.position.rune_address:
            wait_time_in_sec = wait_time_sec_for_rune
        else:
            wait_time_in_sec = wait_time_sec_for_asset

        all_inbound_details = await self.cache.get_inbound_details()
        inbound_details = all_inbound_details.get(asset.chain, None)

        rune_inbound = calc_network_fee(self.cache.native_asset, inbound_details)
        asset_inbound = calc_network_fee(asset, inbound_details)
        rune_outbound = calc_outbound_fee(self.cache.native_asset, inbound_details)
        asset_outbound = calc_outbound_fee(asset, inbound_details)

        # todo cacao?
        total_dust_in_rune, in_asset_fee_in_rune, out_asset_fee_in_rune = await asyncio.gather(
            self.cache.convert(dust_values.asset, self.cache.native_asset),
            self.cache.convert(asset_inbound, self.cache.native_asset),
            self.cache.convert(asset_outbound, self.cache.native_asset),
        )

        memo = THORMemo.withdraw('').build()  # todo

        deposit_amount = CryptoAmount.zero(asset)  # todo!

        return EstimateWithdrawLP(
            can_withdraw=True,
            mode=mode,
            deposit_amount=deposit_amount,
            asset_address=member_detail.position.asset_address,
            rune_address=member_detail.position.rune_address,
            slip_percent=float(slip) * 100.0,
            inbound_fee=LPAmountTotal(
                rune_inbound,
                asset_inbound,
                in_asset_fee_in_rune + rune_inbound
            ),
            inbound_min_to_send=LPAmountTotal(
                dust_values.rune,
                dust_values.asset,
                total_dust_in_rune + dust_values.rune
            ),
            outbound_fee=LPAmountTotal(
                rune_outbound,
                asset_outbound,
                out_asset_fee_in_rune + rune_outbound
            ),
            asset_amount=pool_share.asset,
            rune_amount=pool_share.rune,
            lp_growth=member_detail.lp_growth,
            estimated_wait_seconds=wait_time_in_sec,
            asset_pool=asset_pool.pool.asset,
            errors=[],
            memo=memo,
            inbound_address=inbound_address,
        )

    async def estimate_add_saver(self, add_amount: CryptoAmount) -> EstimateAddSaver:
        """
        Estimate the add liquidity with saver.
        Derived from https://dev.thorchain.org/thorchain-dev/connection-guide/savers-guide

        :param add_amount: CryptoAmount
        :return: EstimateAddSaver
        """
        # check for errors before sending quote
        errors = await self.get_add_savers_estimate_errors(add_amount)

        if errors:
            return EstimateAddSaver.make_error(errors, add_amount.asset)

        # request param amount should always be in 1e8 which is why we pass in adjusted decimals if chain decimals != 8
        if add_amount.amount.decimals != self.native_decimal:
            new_add_amount = get_base_amount_with_diff_decimals(add_amount, self.native_decimal)
        else:
            new_add_amount = add_amount.amount.as_base

        new_add_amount_raw = int(new_add_amount)
        deposit_quote: QuoteSaverDepositResponse = await self.cache.quote_api.quotesaverdeposit(
            asset=str(add_amount.asset),
            amount=new_add_amount_raw,
        )

        if not deposit_quote:
            errors.append(f"Thornode request quote failed")
        if hasattr(deposit_quote, 'error'):
            errors.append(f"Thornode request quote failed: {deposit_quote.error}")

        # The recommended minimum inbound amount for this transaction type & inbound asset.
        # Sending less than this amount could result in failed refunds
        recommended_min_amount_in = int(deposit_quote.recommended_min_amount_in or 0)
        if recommended_min_amount_in and new_add_amount_raw < int(recommended_min_amount_in):
            recommended_in = CryptoAmount.automatic(int(deposit_quote.recommended_min_amount_in), new_add_amount.asset,
                                                    decimals=self.native_decimal)

            errors.append(f"Amount {new_add_amount.amount} is less than recommended min amount {recommended_in}")

        # Error handling
        if errors:
            return EstimateAddSaver.make_error(errors, add_amount.asset)

        # Calculate transaction expiry time of the vault address
        current_date_time = datetime.now()
        minutes_to_add = DEFAULT_EXTRA_ADD_MINUTES
        expiry_date_time = current_date_time + timedelta(minutes=minutes_to_add)

        # Calculate seconds
        if deposit_quote.inbound_confirmation_seconds:
            estimated_wait = deposit_quote.inbound_confirmation_seconds
        else:
            estimated_wait = await self._get_confirmation_counting(add_amount)

        pool_details = await self.cache.get_pool_for_asset(add_amount.asset)
        pool = pool_details.pool

        # Organise fees
        saver_fees = SaverFees(
            affiliate=CryptoAmount.from_base(deposit_quote.fees.affiliate, add_amount.asset),
            asset=add_amount.asset,
            outbound=CryptoAmount.from_base(deposit_quote.fees.outbound, add_amount.asset)
        )

        # Define savers filled capacity
        saver_cap_filled_percent = int(pool.synth_supply) / int(pool.asset_depth) * 100

        # Return object
        return EstimateAddSaver(
            asset_amount=CryptoAmount.from_base(deposit_quote.expected_amount_out, add_amount.asset),
            estimated_deposit_value=CryptoAmount.from_base(deposit_quote.expected_amount_out, add_amount.asset),
            fee=saver_fees,
            expiry=expiry_date_time,
            to_address=deposit_quote.inbound_address,
            memo=deposit_quote.memo,
            estimated_wait_time=estimated_wait,
            can_add_saver=(not errors),
            slip_basis_points=int(deposit_quote.fees.slippage_bps),
            saver_cap_filled_percent=saver_cap_filled_percent,
            errors=errors,
            recommended_min_amount_in=recommended_min_amount_in,
        )

    async def get_add_savers_estimate_errors(self, add_amount: CryptoAmount) -> List[str]:
        errors = []

        if add_amount.amount.internal_amount <= 0:
            errors.append(f'Invalid input amount: {add_amount.amount}; must be greater than 0')

        if add_amount.asset.synth:
            errors.append(f'Cannot add savers for synthetic assets: {add_amount.asset}')

        if add_amount.asset.chain == Chain.THORChain.value:
            errors.append(f'Cannot add RUNE to savers vault')

        pools = await self.cache.get_pools()
        saver_pools = [pool for pool in pools.values() if pool.thornode_details.savers_depth != "0"]
        saver_pool = next((pool for pool in saver_pools if pool.asset == add_amount.asset), None)
        if not saver_pool:
            errors.append(f"{add_amount.asset} does not have a saver's pool")

        inbound_details = await self.cache.get_inbound_details()
        inbound = inbound_details.get(add_amount.asset.chain)
        if inbound is None:
            errors.append(f"no inbound details for chain {add_amount.asset.chain}")
        if inbound.halted_chain:
            errors.append(f"{add_amount.asset.chain} is halted, cannot add")

        pool = pools.get(str(add_amount.asset))
        if pool.pool.status.lower() != pool.AVAILABLE:
            errors.append(f"Pool is not available for this asset {add_amount.asset}")

        # "Quote"-call will check it!
        # inbound_fee = calc_network_fee(add_amount.asset, inbound)
        # if add_amount < inbound_fee:
        #     errors.append(f"Add amount does not cover fees")
        return errors

    async def estimate_withdraw_saver(self,
                                      asset: Asset,
                                      address: str,
                                      withdraw_bps: int,
                                      height: int = 0) -> EstimateWithdrawSaver:
        """
        Estimate the withdrawal liquidity with saver (query THORChain node)
        :param asset: asset to withdraw
        :param address: address to withdraw to
        :param withdraw_bps: basis points to withdraw 0..10k (0..100%)
        :param height: block height to query (optional)
        :return: EstimateWithdrawSaver
        """
        errors = []

        if not asset.chain or not asset.symbol:
            errors.append(f'Invalid asset: {asset}')

        if withdraw_bps < 0 or withdraw_bps > THOR_BASIS_POINT_MAX:
            errors.append(f'Invalid withdraw basis points: {withdraw_bps}; '
                          f'must be between 0 and {THOR_BASIS_POINT_MAX}')

        if asset == self.native_asset or asset.synth:
            errors.append(f"Native Rune and synth assets are not supported only L1's")

        if errors:
            return EstimateWithdrawSaver.make_error(errors, asset)

        try:
            # Request withdraw quote
            withdraw_quote = await self.cache.quote_api.quotesaverwithdraw(
                height=height,
                asset=str(asset),
                address=address,
                withdraw_bps=withdraw_bps
            )
        except ValueError:
            error, withdraw_quote = self._get_error_and_response_from_last_thor_response()
            if error:
                errors.append(error)

            zero = CryptoAmount(Amount.zero(), asset)
            # noinspection PyTypeChecker
            return EstimateWithdrawSaver(
                expected_asset_amount=zero,
                fee=SaverFees(zero, asset, zero),
                expiry=datetime.now(),
                to_address='',
                memo='',
                estimated_wait_time=0,
                slip_basis_points=0,
                dust_amount=zero,
                errors=errors,
                details=withdraw_quote,
            )

        if not withdraw_quote:
            errors.append(f"Thornode request quote failed")
        elif hasattr(withdraw_quote, 'error'):
            errors.append(f"Thornode request quote failed: {withdraw_quote.error}")

        if not withdraw_quote.fees.asset or withdraw_quote.expected_amount_out == '':
            errors.append(f"This address is not found in the savers list of {asset}")

        if errors:
            return EstimateWithdrawSaver.make_error(errors, asset, withdraw_quote)

        # Calculate transaction expiry time of the vault address
        current_date_time = datetime.now()
        minutes_to_add = DEFAULT_EXTRA_ADD_MINUTES
        expiry_date_time = current_date_time + timedelta(minutes=minutes_to_add)
        estimated_wait = int(withdraw_quote.outbound_delay_seconds)
        withdraw_asset = Asset.from_string_exc(withdraw_quote.fees.asset)

        return EstimateWithdrawSaver(
            expected_asset_amount=CryptoAmount.from_base(withdraw_quote.expected_amount_out, asset),
            fee=SaverFees(
                CryptoAmount.from_base(withdraw_quote.fees.affiliate, withdraw_asset),
                withdraw_asset,
                CryptoAmount.from_base(withdraw_quote.fees.outbound, withdraw_asset)
            ),
            expiry=expiry_date_time,
            to_address=withdraw_quote.inbound_address,
            memo=withdraw_quote.memo,
            estimated_wait_time=estimated_wait,
            slip_basis_points=int(withdraw_quote.fees.slippage_bps),
            dust_amount=CryptoAmount.from_base(withdraw_quote.dust_amount, withdraw_asset),
            errors=errors,
            details=withdraw_quote,
        )

    def _get_error_and_response_from_last_thor_response(self) -> Tuple[str, object]:
        try:
            response = self.cache.thornode_client.last_response
            data = response.data
            if isinstance(data, str):
                data = json.loads(data)
            error = data.get('error')
            if not error:
                # if no error message, this means the error has not occurred in the API, but in our code
                raise

            return str(error), response
        except Exception as e:
            return f'Could not pass error info. {e!r}', None

    async def get_saver_position(self, asset: Union[str, Asset], address: str) -> Optional[SaversPosition]:
        """
        Get the position of a saver
        :param asset: asset (pool) to check
        :type asset: Union[str, Asset]
        :param address: saver's address
        :type address: str
        :return: SaversPosition or None if not found
        :rtype: Optional[SaversPosition]
        """
        errors = []

        pool_details = await self.cache.get_pool_for_asset(asset)

        saver = await self.cache.saver_api.saver(str(asset), address)

        if not pool_details or not pool_details.pool:
            errors.append(f"Could not get pool details for {asset}")

        if not saver or not saver.last_add_height:
            errors.append(f"Could not find position for {address}")

        owner_units = int(saver.units)
        last_added = int(saver.last_add_height)
        saver_units = int(pool_details.thornode_details.savers_units)
        asset_depth = int(pool_details.thornode_details.savers_depth)
        redeemable_value = (owner_units / saver_units) * asset_depth
        deposit_amount = CryptoAmount.from_base(saver.asset_deposit_value, asset)
        redeemable_asset_amount = CryptoAmount.from_base(redeemable_value, asset)

        if saver.asset_deposit_value:
            saver_growth = (redeemable_value - saver.asset_deposit_value) / saver.asset_deposit_value * 100.0
        else:
            saver_growth = 0.0

        return SaversPosition(
            deposit_amount,
            redeemable_asset_amount,
            last_added,
            saver_growth,
            errors,
        )

    async def get_loan_quote_open(self, amount: CryptoAmount, target_asset: Asset,
                                  destination: str, min_out: Amount, affiliate_bps: int = 0,
                                  affiliate: str = '', height: int = 0) -> LoanOpenQuote:
        """
        Get a quote for opening a loan
        :param int amount: the asset amount in 1e8 decimals and the asset used to repay the loan
        :param Asset target_asset: the target asset to receive (loan denominated in TOR regardless)
        :param str destination: the destination address, required to generate memo
        :param Amount min_out: the minimum amount of the target asset to accept
        :param str affiliate_bps: the affiliate fee in basis points
        :param int affiliate: the affiliate (address or thorname)
        :param height: optional height (default is the last block)
        :return: LoanOpenQuote
        """
        errors = []

        try:
            resp = await self.cache.quote_api.quoteloanopen(
                asset=str(amount.asset),
                amount=amount.amount.as_base.amount,
                target_asset=str(target_asset),
                destination=destination,
                min_out=min_out.as_base.amount,
                affiliate_bps=int(affiliate_bps),
                affiliate=affiliate,
                height=height
            )
        except ValueError:
            error, _ = self._get_error_and_response_from_last_thor_response()
            if error:
                errors.append(error)
            return LoanOpenQuote.empty_with_errors(errors)

        if resp.recommended_min_amount_in and amount.amount.as_base.amount < resp.recommended_min_amount_in:
            errors.append(f"Amount is less than recommended minimum amount in: {resp.recommended_min_amount_in}")

        if errors:
            return LoanOpenQuote.empty_with_errors(errors)

        return LoanOpenQuote(
            inbound_address=resp.inbound_address,
            expected_wait_time=BlockInformation(
                outbound_delay_blocks=int(resp.expected_delay_blocks),
                outbound_delay_seconds=float(resp.outbound_delay_seconds),
            ),
            fees=QuoteFees(
                asset=Asset.from_string(resp.fees.asset),
                affiliate=int(resp.fees.affiliate),
                outbound=int(resp.fees.outbound),
                total_bps=int(resp.fees.total_bps),
            ),
            slippage_bps=int(resp.slippage_bps),
            router=resp.router,
            expiry=int(resp.expiry),
            warning=resp.warning,
            notes=resp.notes,
            dust_threshold=int(resp.dust_threshold),
            memo=resp.memo,
            expected_amount_out=int(resp.expected_amount_out),
            expected_debt_up=int(resp.expected_debt_up),
            expected_collateral_up=int(resp.expected_collateral_up),
            expected_collateralization_ratio=float(resp.expected_collateralization_ratio),
            errors=errors,
            recommended_min_amount_in=int(resp.recommended_min_amount_in),
        )

    async def get_loan_quote_close(self, amount: CryptoAmount, from_address: str,
                                   loan_asset: Asset, loan_owner: str,
                                   min_out: Amount, height: int = 0) -> LoanCloseQuote:
        """
        Get a quote for closing a loan.

        :param amount:
            CryptoAmount object representing the asset amount in 1e8 decimals and the asset used to repay the loan.
        :param from_address:
            The address that is paying off the loan.
        :param loan_asset:
            Asset object representing the collateral asset of the loan.
        :param loan_owner:
            The owner of the loan collateral.
        :param min_out:
            Amount object representing the minimal threshold for output amount.
        :param height:
            Optional block height, defaults to the current tip if not provided.
        :type height: int, optional
        :return:
            LoanCloseQuote object representing the quote for closing the loan.
        :rtype: LoanCloseQuote
        """
        errors = []

        try:
            resp = await self.cache.quote_api.quoteloanclose(
                height=height,
                asset=str(amount.asset),
                amount=amount.amount.as_base.amount,
                from_address=from_address,
                loan_asset=str(loan_asset),
                loan_owner=loan_owner,
                min_out=min_out.as_base.amount
            )
        except ValueError:
            error, _ = self._get_error_and_response_from_last_thor_response()
            if error:
                errors.append(error)
            return LoanCloseQuote.empty_with_errors(errors)

        if errors:
            return LoanCloseQuote.empty_with_errors(errors)

        return LoanCloseQuote(
            inbound_address=resp.inbound_address,
            expected_wait_time=BlockInformation(
                outbound_delay_blocks=int(resp.outbound_delay_blocks),
                outbound_delay_seconds=float(resp.outbound_delay_seconds),
            ),
            fees=QuoteFees(
                asset=Asset.from_string(resp.fees.asset),
                liquidity=int(resp.fees.liquidity),
                outbound=int(resp.fees.outbound),
                total_bps=int(resp.fees.total_bps),
            ),
            slippage_bps=int(resp.slippage_bps),
            router=resp.router,
            expiry=int(resp.expiry),
            warning=resp.warning,
            notes=resp.notes,
            dust_threshold=int(resp.dust_threshold),
            memo=resp.memo,
            expected_amount_out=int(resp.expected_amount_out),
            expected_collateral_down=int(resp.expected_collateral_down),
            expected_debt_down=int(resp.expected_debt_down),
            errors=errors,
            recommended_min_amount_in=int(resp.recommended_min_amount_in),
        )

    async def estimate_thor_name(self, is_update: bool, thorname: str, expiry: datetime = None) -> THORNameEstimate:
        """
        Estimate the cost of registering or updating a THORName. Calculations are worked out locally based on the
        current network values and the current state of the THORName.

        :param is_update: bool - True if updating an existing THORName
        :param thorname: str - the THORName to register or update
        :param expiry: datetime - the desired expiry date of the THORName
        :return: THORNameEstimate
        """
        thor_name_details = await self.cache.get_name_details(thorname)
        if thor_name_details and thor_name_details.owner != '' and not is_update:
            return THORNameEstimate.error("THORName already registered; cannot register again", 0)

        if not thor_name_details and is_update:
            return THORNameEstimate.error("THORName not found; cannot update", 0)

        current_thorchain_height = await self.cache.get_native_block_height()

        current_expire = int(thor_name_details.expire) if thor_name_details else 0

        current_height_for_expiry = current_expire if is_update else current_thorchain_height

        blocks_to_add_to_expiry = 0 if is_update else THORNAME_BLOCKS_ONE_YEAR

        if expiry:
            current_timestamp = datetime.now().timestamp()
            expiry_timestamp = expiry.timestamp()
            interval_to_expire = expiry_timestamp - current_timestamp
            blocks_to_expire = int(interval_to_expire / self.native_block_time)
            new_height_for_expiry = current_thorchain_height + blocks_to_expire
            if current_expire:
                blocks_to_add_to_expiry = new_height_for_expiry - current_height_for_expiry
            else:
                blocks_to_add_to_expiry = blocks_to_expire

        # compute value
        constants = await self.cache.get_network_values()
        one_time_fee = Amount.zero(self.native_decimal) if is_update else Amount.from_base(
            constants.get(Mimir.TNS_REGISTER_FEE, 0), self.native_decimal)
        fee_per_block = constants.get(Mimir.TNS_FEE_PER_BLOCK, 0)
        total_fee_per_block = Amount.from_base(
            fee_per_block * max(blocks_to_add_to_expiry, 0),
            self.native_decimal
        )
        total_cost_amount = one_time_fee.as_asset + total_fee_per_block.as_asset
        total_cost = CryptoAmount(total_cost_amount, self.native_asset)

        return THORNameEstimate(
            can_register=True,
            reason='',
            cost=total_cost,
            details=thor_name_details,
            last_block_number=current_thorchain_height
        )

    async def get_inbound_details(self, chain: Union[str, Chain]) -> InboundDetail:
        """
        Get the inbound details for a given chain. It contains the recommended fee, fee units,
        router and vault addresses and the status of the chain.

        :param chain: str or Chain enum value
        :return: InboundDetail
        """
        return await self.cache.get_details_for_chain(chain)

    def tracker(self,
                wallet=None, inbound_chain: Optional[Chain] = None, outbound_chain: Optional[Chain] = None,
                inbound_client: Optional[XChainClient] = None,
                outbound_client: Optional[XChainClient] = None) -> TransactionTracker:
        """
        Create a transaction tracker for a given wallet or chain/client combination.
        Provide either a wallet with inbound/outbound chain idents or clients combination, not both.

        :param wallet: Optional[Wallet] - the wallet to get the client from
        :param inbound_chain: Optional[Chain] - the inbound Chain enum item
        :param outbound_chain: Optional[Chain] - the outbound Chain enum item
        :param inbound_client: Optional[XChainClient] - directly provide the inbound client
        :param outbound_client: Optional[XChainClient] - directly provide the outbound client
        :return: TransactionTracker object
        """

        if (inbound_client or outbound_client) and (wallet and (inbound_chain or outbound_chain)):
            raise ValueError("Cannot provide both wallet and chain/client")

        if not inbound_client and wallet:
            inbound_client = wallet.get_client(inbound_chain)
            if not inbound_client:
                raise ValueError(
                    f"You provided a wallet but no client for inbound chain {inbound_chain} could be found")

        if not outbound_chain and wallet:
            outbound_client = wallet.get_client(outbound_chain)
            if not outbound_client:
                raise ValueError(
                    f"You provided a wallet but no client for outbound chain {outbound_chain} could be found")

        return TransactionTracker(self.cache, self.chain_attributes,
                                  inbound_chain_client=inbound_client,
                                  outbound_chain_client=outbound_client)



import datetime

import pytest
import requests_mock

from xchainpy2_thorchain_query import THORChainQuery


@pytest.mark.asyncio
async def test_estimate_thor_name():
    with requests_mock.Mocker() as m:
        # m.get(f'{URLs.THORNode.MAINNET}/thorchain/inbound_addresses', json=load_json('inbound_addresses'))

        query = THORChainQuery()

        r = await query.estimate_thor_name(False, 'fefefwfebfghh')
        assert r.can_register
        # 10 rune once + 1 rune / year
        assert 11 < float(r.cost.amount) < 12

        r = await query.estimate_thor_name(False, 'fefefwfebfghh',
                                           expiry=datetime.datetime.now() + datetime.timedelta(days=365 * 3))
        assert r.can_register
        # 10 rune once + 1 rune / year => 13
        assert 13 < float(r.cost.amount) < 13.5

        r = await query.estimate_thor_name(True, 'fefefwfebfghh')
        assert not r.can_register

        # already registered, cannot register again

        r = await query.estimate_thor_name(False, 't')
        assert not r.can_register

        await query.close()



class URLs:
    """
    A collection of URLs for public THORChain APIs (THORNode, RPC, Midgard).
    See https://ops.ninerealms.com/links for more links.
    """

    class THORNode:
        """
        THORNode API URLs.
        """
        PUBLIC = 'https://thornode.thorchain.info'
        NINE_REALMS = 'https://thornode.ninerealms.com'
        THORSWAP = 'https://thornode.thorswap.net'

        # Pre-hard-fork endpoints (blocks 1 through 4786559):
        ARCHIVE_V0 = 'https://thornode-v0.ninerealms.com'
        # Post-hard-fork endpoints (block 4786560 to present):
        ARCHIVE_V1 = 'https://thornode-v1.ninerealms.com'

        MAINNET = NINE_REALMS
        STAGENET = 'https://stagenet-thornode.ninerealms.com'
        TESTNET = 'https://testnet.thornode.thorchain.info'

    class RPC:
        """
        THORChain RPC URLs.
        """
        PUBLIC = 'https://rpc.thorchain.info'
        NINE_REALMS = 'https://rpc.ninerealms.com'
        THORSWAP = 'https://rpc.thorswap.net'

        ARCHIVE_V0 = 'https://rpc-v0.ninerealms.com'

        MAINNET = NINE_REALMS
        STAGENET = 'https://stagenet-rpc.ninerealms.com'
        TESTNET = 'https://testnet.rpc.thorchain.info/'

    class Midgard:
        """
        Midgard API URLs.
        """
        PUBLIC = 'https://midgard.thorchain.info'
        NINE_REALMS = 'https://midgard.ninerealms.com'
        THORSWAP = 'https://midgard.thorswap.net'

        MAINNET = NINE_REALMS
        STAGENET = 'https://stagenet-midgard.ninerealms.com'
        TESTNET = 'https://testnet.midgard.thorchain.info'



import pytest
import requests_mock

from .common import *
from xchainpy2_thorchain_query import THORChainCache


@pytest.mark.asyncio
async def test_timeout_thornode(thornode_instance):
    with requests_mock.Mocker() as m:
        m.get(f'{URLs.THORNode.MAINNET}/thorchain/inbound_addresses', json=load_json('inbound_addresses'))

        # Client timeout is 0.01 seconds (very small)

        thornode_instance.configuration.timeout = 0.01

        cache = THORChainCache(None, thornode_instance)

        # This should fail with a timeout error
        with pytest.raises(Exception) as e:
            await cache.get_inbound_details()
        assert 'All backup hosts failed' in str(e.value)
        assert e.value.__cause__.__class__.__name__ == 'TimeoutError'

        thornode_instance.configuration.timeout = 99

        d = await cache.get_inbound_details()
        assert len(d) > 0

        await cache.close()


@pytest.mark.asyncio
async def test_timeout_midgard(thornode_instance, midgard_instance):
    with requests_mock.Mocker() as m:
        m.get(f'{URLs.Midgard.MAINNET}/v2/thorchain/pools', json=load_json('midgard_pools'))
        m.get(f'{URLs.THORNode.MAINNET}/thorchain/inbound_addresses', json=load_json('inbound_addresses'))

        # Client timeout is 0.01 seconds (very small)

        midgard_instance.configuration.timeout = 0.01
        thornode_instance.configuration.timeout = 0.01

        cache = THORChainCache(midgard_instance, thornode_instance)

        # This should fail with a timeout error
        with pytest.raises(Exception) as e:
            await cache.get_pools()
        assert 'All backup hosts failed' in str(e.value)
        assert e.value.__cause__.__class__.__name__ == 'TimeoutError'

        midgard_instance.configuration.timeout = 99
        thornode_instance.configuration.timeout = 99

        d = await cache.get_pools()
        assert len(d) > 0

        await cache.close()
